# Copyright 2023 The T5X Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for partitioning."""

import abc
import collections
import dataclasses
import functools
import typing
from typing import Any, Callable, Optional, Sequence, Set, Tuple, Union

import jax
import numpy as np
from absl import logging
from flax import traverse_util
from flax.linen import partitioning as flax_partitioning
from jax import numpy as jnp
from jax import random
from jax.experimental import multihost_utils
from jax.experimental.mesh_utils import create_hybrid_device_mesh
from jax.experimental.pjit import pjit
from jax.sharding import Mesh, PartitionSpec

JaxDevice = jax.Device
TpuMesh = Tuple[int, int, int, int]  # (x, y, z, num_cores).
OtherMesh = Tuple[int, int]
HardwareMesh = Union[TpuMesh, OtherMesh]
LogicalAxisRules = Sequence[Tuple[str, Optional[str]]]

cached_property = property


class AxisNames(tuple):
	"""Tuple of strings specifying name for each axis.

	We create a separate class for this so JAX's pytree utilities can distinguish
	it from a tuple that should be treated as a pytree, instead treating it as a
	leaf.
	"""

	def __new__(cls, *names):
		return tuple.__new__(AxisNames, names)

	def __repr__(self):
		return "AxisNames%s" % tuple.__repr__(self)


def with_sharding_constraint(x, axis_resources):
	"""Wrapper for lax.with_sharding_constraint, no-op on cpu or outside pjit."""
	if jax.devices()[0].platform == "cpu" or not global_mesh_defined():
		return x
	else:
		return jax.lax.with_sharding_constraint(x, axis_resources)


def bounds_from_last_device(last_device: jax.Device) -> HardwareMesh:
	"""Get the bound from the given last device."""
	# Must be passed the device at the highest-coordinate corner of the
	# relevant mesh, which is a requirement we know is satisfied by the last
	# device in jax.devices().
	if hasattr(last_device, "coords"):
		x, y, z = last_device.coords
		return x + 1, y + 1, z + 1, last_device.core_on_chip + 1
	else:
		# On non-TPU platforms, the "mesh" is hosts x devices per host in order
		# to take advantage of faster within-host interconnect.
		return jax.process_count(), jax.local_device_count()


def get_coords(device: jax.Device) -> HardwareMesh:
	"""Returns the coordinates of the given device."""
	if hasattr(device, "coords"):
		return *device.coords, device.core_on_chip
	return device.process_index, device.id % jax.local_device_count()


def global_mesh_defined():
	"""Checks if global xmap/pjit mesh resource environment is defined."""
	maps_env = jax.experimental.maps.thread_resources.env
	return maps_env.physical_mesh.devices.shape != ()  # pylint: disable=g-explicit-bool-comparison


def get_mesh(
	model_parallel_submesh: HardwareMesh,
	input_devices: Sequence[JaxDevice] = (),
	input_local_devices: Sequence[JaxDevice] = (),
	tile_by_host_if_needed: bool = True,
	backend: Optional[str] = None,
) -> Mesh:
	"""Construct an xmap/pjit Mesh for the given model-parallel submesh.

	The resulting mesh has two resource axes: 'model', with the provided submesh
	shape, and 'data', which covers the rest of the mesh.

	Args:
	  model_parallel_submesh: a HardwareMesh spec, namely (x,y,z,core) on TPU for
	    a single model-parallel replica's "tile" in the physical device mesh. The
	    first three elements (`x`, `y`, and `z`) should be factors of the pod
	    slice; e.g., if you are using df_4x8, then `x` should be a factor of 4
	    (one of 1, 2, 4), `y` should be a factor of 8 (one of 1, 2, 4, 8), and `z`
	    must be 1, because TPU v3 slices are only 2D. `z` can be >1 for TPU v4
	    (and maybe later TPUs) that allow 3D slices. `core` is the number of cores
	    to use from each TPU node. As communication is usually fastest inside the
	    same node, if you need a tile of more than 1 core, then
	    you should first increase `core`: e.g., for TPU v3, (1,1,1,2) is better
	      than (2,1,1,1). To pick a good spec, try a few possible values until you
	      get high TPU utilization.
	  input_devices: the devices to use, will use jax.devices() if this is not
	    set.
	  input_local_devices: the local devices to use, will use jax.local_devices()
	    if this is not set.
	  tile_by_host_if_needed: JAX currently requires that the parts of any sharded
	    array that are located on one host's local devices form a single
	    contiguous slice. A best effort will be made to achieve this without
	    "tiling" the device assignment over hosts (which can reduce XLA collective
	    performance). If this flag is True, then the device assignment will be
	    tiled over hosts if necessary to satisfy this constraint and create a
	    buildable mesh; if false, mesh construction will fail instead.
	  backend: get devices from the pinned backend, if specified. This is
	    useful for explicitly specifying the devices other than relying on
	    jax_platform_name.

	Returns:
	  A xmap / pjit Mesh containing the virtual device mesh with data, model axes.
	"""
	input_devices = input_devices or jax.devices(backend)
	input_local_devices = input_local_devices or jax.local_devices(0, backend)
	# Sort input_devices based on coords, as backends might not return devices
	# in order.
	last_device = sorted(input_devices, key=get_coords)[-1]
	last_input_local_devices = sorted(input_local_devices, key=get_coords)[-1]
	logging.info(
		"last device coords : %r\nlast local device coords: %r",
		get_coords(last_device),
		get_coords(last_input_local_devices),
	)
	global_hardware_mesh = bounds_from_last_device(last_device)
	mesh_ndim = len(global_hardware_mesh)
	local_hardware_mesh = bounds_from_last_device(last_input_local_devices)
	mesh_err = (
		f"each dimension of the model parallel submesh {model_parallel_submesh} "
		"must be a factor of the corresponding dimension of the global device "
		f"mesh {global_hardware_mesh}"
	)
	assert not any(
		g % m
		for g, m in zip(global_hardware_mesh, model_parallel_submesh)  # noqa:B905
	), mesh_err
	assert not any(g % l for g, l in zip(global_hardware_mesh, local_hardware_mesh))  # noqa:B905
	devices = np.empty(global_hardware_mesh, dtype=object)
	for device in input_devices:
		device_coords = get_coords(device)
		devices[device_coords] = device
	tile_by_host = tile_by_host_if_needed
	if len(global_hardware_mesh) == 4:
		# enable contiguous local chunks without host tiling by making Z major
		global_hardware_mesh = typing.cast(Tuple[int, int, int, int], global_hardware_mesh)
		model_parallel_submesh = typing.cast(
			Tuple[int, int, int, int], model_parallel_submesh
		)
		gx, gy, gz, gc = global_hardware_mesh
		mx, my, mz, mc = model_parallel_submesh
		if (mx == gx > 1 and my == mz == 1) or (mx == 1 and my == gy > 1 and mz == gz > 1):
			logging.info("ensuring YZ plane has a Z-major device order")
			# YZ should be ZY
			assert mc == gc, (mc, gc)
			global_hardware_mesh = gx, gz, gy, gc
			model_parallel_submesh = mx, mz, my, mc
			devices = devices.swapaxes(1, 2)
			tile_by_host = False
		if (my == gy > 1 and mx == mz == 1) or (my == 1 and mx == gx > 1 and mz == gz > 1):
			logging.info("ensuring XZ plane has a Z-major device order")
			# XZ should be ZX
			assert mc == gc, (mc, gc)
			global_hardware_mesh = gz, gy, gx, gc
			model_parallel_submesh = mz, my, mx, mc
			devices = devices.swapaxes(0, 2)
			tile_by_host = False
	if tile_by_host:
		logging.warning(
			"Tiling device assignment mesh by hosts, which may lead to "
			"reduced XLA collective performance. To avoid this, modify "
			"the model parallel submesh or run with more tasks per host."
		)
		tile_err = (
			"to tile the mesh by hosts, each dimension of the model parallel "
			"submesh must be either a factor or a multiple of the corresponding "
			"dimension of the per-host submesh"
		)

		def dh_dd_mh_md(g: int, m: int, l: int) -> Tuple[int, int, int, int]:
			"""Split a global mesh dimension into four tiling components.

			Args:
			  g: global mesh bounds dimension size
			  m: model-parallel submesh bounds dimension size
			  l: local submesh bounds dimension size

			Returns:
			  The resulting tuple divides the dimension into the hosts component of
			  the data-parallel submesh, the devices component of the data-parallel
			  submesh, the hosts component of the model-parallel submesh, and the
			  devices component of the model-parallel submesh.
			"""
			d = g // m
			if m >= l:
				assert not m % l, tile_err
				return d, 1, m // l, l
			else:
				assert not l % m, tile_err
				return d // (l // m), l // m, 1, m

		dh_dd_mh_md_tups = map(
			dh_dd_mh_md,
			global_hardware_mesh,
			model_parallel_submesh,
			local_hardware_mesh,
		)
		devices = devices.reshape(*(s for t in dh_dd_mh_md_tups for s in t))
		devices = devices.transpose(
			*(4 * i for i in range(mesh_ndim)),
			*(4 * i + 1 for i in range(mesh_ndim)),
			*(4 * i + 2 for i in range(mesh_ndim)),
			*(4 * i + 3 for i in range(mesh_ndim)),
		)
	else:
		model_data_tups = [
			(g // m, m)
			for g, m in zip(global_hardware_mesh, model_parallel_submesh)  # noqa:B905
		]
		devices = devices.reshape(*(s for t in model_data_tups for s in t))
		devices = devices.transpose(
			*(2 * i for i in range(mesh_ndim)), *(2 * i + 1 for i in range(mesh_ndim))
		)
	# reshape to (data, model)
	devices = devices.reshape(-1, np.prod(model_parallel_submesh))
	global_mesh = Mesh(devices, ["data", "model"])
	logging.info("global_mesh axis_names: %s", global_mesh.axis_names)
	logging.info("global_mesh devices: %s", global_mesh.devices)
	logging.info("global_mesh devices shape: %s", global_mesh.devices.shape)
	return global_mesh


def get_cpu_mesh() -> Mesh:
	"""Trivial mesh for CPU Testing."""
	devices = np.empty((jax.process_count(), jax.local_device_count()), dtype=object)
	for device in jax.devices():
		devices[device.process_index, device.id % jax.local_device_count()] = device
	return Mesh(devices, ["data", "model"])


def get_gpu_mesh(num_partitions: int) -> Mesh:
	"""Mesh for GPUs that preferentially places 'model' on NVLink."""
	nvlink_size = jax.local_device_count()
	dcn_size = jax.process_count()
	nvlink_mp = min(num_partitions, nvlink_size)
	nvlink_dp, extra1 = divmod(nvlink_size, nvlink_mp)
	dcn_mp, extra2 = divmod(num_partitions, nvlink_mp)
	assert not (extra1 or extra2), (
		"number of partitions on GPU must be a factor"
		" or multiple of the number of local devices"
	)
	dcn_dp = dcn_size // dcn_mp

	devices = create_hybrid_device_mesh(
		mesh_shape=[nvlink_dp, nvlink_mp],
		dcn_mesh_shape=[dcn_dp, dcn_mp],
		process_is_granule=True,
	)

	global_mesh = Mesh(devices, ["data", "model"])
	logging.info("global_mesh axis_names: %s", global_mesh.axis_names)
	logging.info("global_mesh devices: %s", global_mesh.devices)
	return global_mesh


def default_mesh(
	num_partitions: int,
	model_parallel_submesh: Optional[HardwareMesh] = None,
	backend: Optional[str] = None,
	ici_mesh_shape: Optional[HardwareMesh] = None,
	dcn_mesh_shape: Optional[HardwareMesh] = None,
) -> Mesh:
	"""Attempt to return a default mesh for simple cases.

	Args:
	  num_partitions: number of partitions to use, will be ignored if
	    model_parallel_submesh is provided.
	  model_parallel_submesh: 4-tuple that specifies the x,y,z,c submesh to use as
	    the model-parallel device tile.
	  backend: get devices from the pinned backend, if specified. This is useful
	    for explicitly specifying the devices other than relying on
	    jax_platform_name.
	  ici_mesh_shape: Shape of the logical mesh used for SPMD parallelism in each
	    slice. The meaning of each mesh axis is defined by mesh_axis_names, so
	    these two params must be the same length. If dcn_mesh_shape is present,
	    the overall mesh is the product of ici_mesh_shape and dcn_mesh_shape. For
	    example, an ici_mesh_shape of [2, 3, 4] with mesh_axis_names ['replica',
	    'data', 'model'] indicates 2-way replica parallelism, 3-way data
	    parallelism, and 4-way model parallelism over 24 devices. None, the
	    default, is equivalent to a sequence of ones and means that the model is
	    placed on a single device.
	  dcn_mesh_shape: Shape of the logical mesh used for SPMD parallelism over
	    multiple slices. The overall mesh is the product of ici_mesh_shape and
	    dcn_mesh_shape, and the meaning of each mesh axis is defined by
	    mesh_axis_names, so these three params must be the same length.

	Returns:
	  xmap/pjit 2D Mesh with 'data', 'model' mesh axes if single-slice, otherwise
	  3D Mesh with 'replica', 'data', and 'model' mesh axes.
	"""
	devices = jax.devices(backend)
	last_device = devices[-1]
	platform = last_device.platform
	device_kind = last_device.device_kind
	bounds = bounds_from_last_device(last_device)

	if ici_mesh_shape is not None and dcn_mesh_shape is not None:
		device_mesh = create_hybrid_device_mesh(
			ici_mesh_shape,
			dcn_mesh_shape,
			devices=devices,
		)
		multi_slice_global_mesh = Mesh(device_mesh, ["replica", "data", "model"])
		logging.info(
			"multi_slice_global_mesh axis_names: %s",
			multi_slice_global_mesh.axis_names,
		)
		logging.info("multi_slice_global_mesh devices: %s", multi_slice_global_mesh.devices)
		logging.info(
			"multi_slice_global_mesh devices shape: %s",
			multi_slice_global_mesh.devices.shape,
		)
		return multi_slice_global_mesh

	if model_parallel_submesh:
		return get_mesh(model_parallel_submesh, backend=backend)

	if platform == "cpu":
		return get_cpu_mesh()
	elif platform == "gpu":
		return get_gpu_mesh(num_partitions)

	mps = None
	if device_kind in ("TPU v2", "TPU v3"):
		if num_partitions == 1:
			mps = (1, 1, 1, 1)
		elif num_partitions == 2:
			mps = (1, 1, 1, 2)
		elif num_partitions == 4:
			mps = (2, 1, 1, 2)
		elif num_partitions == 8:
			mps = (2, 2, 1, 2)
		elif num_partitions == 16:
			mps = (4, 2, 1, 2)
	# assume the use of megacore on TPU v4
	elif (device_kind == "TPU v4" or device_kind == "TPU v4 lite") and bounds[3] == 1:
		if num_partitions == 1:
			mps = (1, 1, 1, 1)
		elif num_partitions == 2:
			mps = (1, 2, 1, 1)
		elif num_partitions == 4:
			if bounds[0] >= 4:
				mps = (4, 1, 1, 1)
			else:
				mps = (2, 2, 1, 1)
		elif num_partitions == 8:
			if bounds[2] >= 8:
				mps = (1, 1, 8, 1)
			else:
				mps = (4, 2, 1, 1)
		elif num_partitions == 16:
			if bounds[2] >= 16:
				mps = (1, 1, 16, 1)
			elif bounds[0] >= 8:
				mps = (8, 2, 1, 1)
			elif bounds[0] >= 4:
				mps = (4, 4, 1, 1)
			else:
				mps = (2, 2, 4, 1)

	if mps is None:
		raise ValueError(
			"No default mesh for this configuration: specify "
			"config.model_parallel_submesh explicitly. \n"
			f"Platform: {platform}\n"
			f"Device kind: {device_kind}\n"
			f"Num partitions: {num_partitions}\n"
			f"Bounds: {bounds}"
		)
	return get_mesh(mps, backend=backend)


# Data chunking helper.
# -----------------------------------------------------------------------------
@dataclasses.dataclass
class LocalChunkInfo:
	# The logical slice of an array located on this host's local devices.
	slice: Tuple[slice, ...]
	# A unique index for this host/local chunk among chunks with the same slice.
	replica_id: int


class LocalChunker:
	"""Utility class to aid chunking of sharded arrays in multihost settings."""

	def __init__(self, global_mesh: Mesh):
		self.global_mesh = global_mesh
		local_mesh = global_mesh.local_mesh
		first_local_device = local_mesh.devices.reshape(-1)[0]
		host_location = collections.OrderedDict(
			zip(  # noqa:B905
				global_mesh.shape.keys(),
				list(zip(*np.nonzero(global_mesh.devices == first_local_device)))[0],  # noqa:B905
			)
		)
		self.num_chunks = collections.OrderedDict()
		self.chunk_ids = collections.OrderedDict()
		self.mesh_axes = list(global_mesh.shape.keys())
		for mesh_axis in self.mesh_axes:
			num_devices_per_chunk = local_mesh.shape[mesh_axis]
			self.num_chunks[mesh_axis] = global_mesh.shape[mesh_axis] // num_devices_per_chunk
			self.chunk_ids[mesh_axis] = host_location[mesh_axis] // num_devices_per_chunk

	def get_local_chunk_info(
		self, global_shape: Tuple[int, ...], mesh_axes: Sequence[Optional[str]]
	) -> LocalChunkInfo:
		"""Get the local chunk info for a given array shape and sharded axes.

		Args:
		  global_shape: the global, unsharded shape of the array to chunk.
		  mesh_axes: a sequence of names (or None) of equal rank to `global_shape`
		    that specifies which mesh dimensions the array is sharded along.

		Returns:
		  LocalChunkInfo containing the logical slices of the array found on this
		  host's local devices, as well as the replica index for this chunk among
		  chunks with the same slice. The latter is used to determine which
		  host should write this chunk during checkpointing.
		"""
		local_slice = [slice(None) for dim in global_shape]
		sharded_mesh_axes = set()
		for i, (mesh_axis, size) in enumerate(zip(mesh_axes, global_shape)):  # noqa:B905
			if not mesh_axis:
				continue
			sharded_mesh_axes.add(mesh_axis)
			if not isinstance(mesh_axis, str):
				raise NotImplementedError()
			chunk_id = self.chunk_ids[mesh_axis]
			chunk_size = size // self.num_chunks[mesh_axis]
			local_slice[i] = slice(chunk_id * chunk_size, (chunk_id + 1) * chunk_size)

		replica_id = self.get_replica_id(sharded_mesh_axes)

		return LocalChunkInfo(tuple(local_slice), replica_id)

	def get_shard_id(self, sharded_mesh_axes: Union[str, Set[Optional[str]]]) -> int:
		"""Given mesh axes used for sharding, computes current host's shard id.

		To give an example, let's say there are two axes globally: replica, data,
		and model, the mesh axes for sharding is ('replica', 'data'), which means we
		are going to partition an array along 'replica' and 'data' axes.
		The shard_id is to show the index of the current local host along the
		sharding axes (in this example, it's 'replica' and 'data' axes).

		More concretely, let's say we have 4 local hosts, and we use 'replica' and
		'data' axes for data parallel (2 hosts along the replica axis, and 2 host
		along the data axis). The host located in ('replica': 0, 'data': 0), we
		should assign data shard-0 to it. For host ('replica': 0, 'data': 1), we
		assign shard-1. For host ('replica': 1, 'data': 0), we assign shard-2.
		For host ('replica': 1, 'data': 1), we assign shard-3.

		Note: the host location along 'replica' and 'data' axes, e.g.,
		('replica': 0, 'data': 0) is named chunk_id and stored in
		self._local_chunker.chunk_ids[axis].

		Args:
		  sharded_mesh_axes: the mesh axes for sharding.

		Returns:
		  the index of the current local host along the sharding axes.
		"""
		if isinstance(sharded_mesh_axes, str):
			sharded_mesh_axes = (sharded_mesh_axes,)

		shard_id = 0
		for mesh_axis in sharded_mesh_axes:
			chunk_id = self.chunk_ids[mesh_axis]
			shard_id = shard_id * self.num_chunks[mesh_axis] + chunk_id

		return shard_id

	def get_replica_id(self, sharded_mesh_axes: Union[str, Set[Optional[str]]]) -> int:
		"""Given mesh axes used for sharding, computes current host's replica id.

		To give an example, let's say there are two axes globally: data, and model,
		the mesh axes for sharding is ('data', ), which means we are going to
		partition an array along 'data' axis and replicate it along 'model' axis.
		The replica_id is to show the index of the current local host along the
		'model' axis.

		Args:
		  sharded_mesh_axes: the mesh axes for sharding.

		Returns:
		  the index of the current local host along the non-sharding axes (i.e.,
		  replicating axes).
		"""
		if isinstance(sharded_mesh_axes, str):
			sharded_mesh_axes = (sharded_mesh_axes,)

		replicated_mesh_axes = [
			mesh_axis for mesh_axis in self.mesh_axes if mesh_axis not in sharded_mesh_axes
		]
		replica_id = 0
		for mesh_axis in replicated_mesh_axes:
			chunk_id = self.chunk_ids[mesh_axis]
			replica_id = replica_id * self.num_chunks[mesh_axis] + chunk_id

		return replica_id


def standard_logical_axis_rules(
	activation_partitioning_dims: int = 1,
	parameter_partitioning_dims: int = 1,
	additional_rules: Optional[LogicalAxisRules] = None,
) -> LogicalAxisRules:
	"""Default sharding rules for T5X model in terms of logical axis names.

	Args:
	  activation_partitioning_dims: enables 2-D activation sharding when set to 2.
	  parameter_partitioning_dims: enables 2-D parameter sharding when set to 2.
	  additional_rules: additional rules (a sequence of tuples) that will be
	    appended to the standard rules.

	Returns:
	  Sequence of logical axis rules
	"""
	logging.info(
		"`activation_partitioning_dims` = %d, `parameter_partitioning_dims` = %d",
		activation_partitioning_dims,
		parameter_partitioning_dims,
	)

	if activation_partitioning_dims == 1 and parameter_partitioning_dims == 1:
		rules = [
			("batch", "data"),
			("vocab", "model"),
			("embed", None),
			("mlp", "model"),
			("heads", "model"),
			("kv", None),
			("joined_kv", "model"),  # joined heads+kv dim in 2D attn param layouts
		]
	elif activation_partitioning_dims == 2 and parameter_partitioning_dims == 1:
		rules = [
			("batch", "data"),
			("vocab", "model"),
			("mlp", "model"),
			("heads", "model"),
			("kv", None),
			("joined_kv", "model"),
			("embed", "model"),
		]
	elif activation_partitioning_dims == 1 and parameter_partitioning_dims == 2:
		rules = [
			("batch", "data"),
			("vocab", "model"),
			("mlp", "model"),
			("heads", "model"),
			("kv", None),
			("joined_kv", "model"),
			("embed", "data"),
		]
	elif activation_partitioning_dims == 2 and parameter_partitioning_dims == 2:
		rules = [
			("batch", "data"),
			("vocab", "model"),
			("mlp", "model"),
			("heads", "model"),
			("kv", None),
			("joined_kv", "model"),
			("embed", "model"),
			("embed", "data"),
		]
	else:
		raise ValueError(
			f"`activation_partitioning_dims` = {activation_partitioning_dims} "
			f"`parameter_partitioning_dims` = {parameter_partitioning_dims} "
			"is not supported."
		)

	# Add the common rules for the replicated logical axes names.
	replicated_rules = [
		("relpos_buckets", None),
		("abspos_buckets", None),
		("length", None),
		("layers", None),
		("stack", None),
		("mlp_activations", None),
	]
	rules.extend(replicated_rules)

	if additional_rules:
		rules.extend(additional_rules)

	return rules


# NB: This needs to be top-level for the jax compilation cache.
def _id_fn(x, ix):
	"""Identity function for copying parameters to the devices, sharded."""
	# A pure identity such as `lambda x, *: x` can get optimized away, so we
	# include a random.split as a cheap function that cannot be optimized away.
	y = random.split(random.PRNGKey(jnp.array(ix, dtype=jnp.uint32)))
	return x, y


@dataclasses.dataclass
class DataLayout:
	"""Represents data layout for the partitioned model."""

	batch_size: int
	shard_id: int
	num_shards: int
	is_first_host_in_replica_set: bool


PartitionedCallable = Callable[..., Any]
CompiledPartitionedCallable = Callable[..., Any]


class BasePartitioner(metaclass=abc.ABCMeta):
	"""Interface for partitioning computations across hardware devices."""

	def __init__(
		self,
		num_partitions: Optional[int] = None,
		model_parallel_submesh: Optional[HardwareMesh] = None,
		params_on_devices: bool = True,
		backend: Optional[str] = None,
		ici_mesh_shape: Optional[HardwareMesh] = None,
		dcn_mesh_shape: Optional[HardwareMesh] = None,
	):
		"""Configures the partitioner.

		Args:
		  num_partitions: the number of partitions to use. Ignored if
		    `model_parallel_submesh` is provided.
		  model_parallel_submesh: 4-tuple that specifies the x,y,z,c submesh to use
		    as the model-parallel device tile. This submesh is used for the larger
		    of the two parameter dimensions, and, if 2-D activation sharding is
		    enabled, for the model dimension of activations. The rest of the mesh is
		    used for data parallelism and, if 2-D parameter sharding is enabled, the
		    other parameter dimension.
		  params_on_devices: whether to keep the params on devices, if False -
		    params stay in the host memory. Note that some partitioners might ignore
		    this setting, for example if they don't support storing all params on
		    device memory.
		  backend: get devices from the pinned backend, if specified. This is useful
		    for explicitly specifying the devices other than relying on
		    jax_platform_name.
		  ici_mesh_shape: Shape of the logical mesh used for SPMD parallelism in
		    each slice. The meaning of each mesh axis is defined by mesh_axis_names,
		    so these two params must be the same length. If dcn_mesh_shape is
		    present, the overall mesh is the product of ici_mesh_shape and
		    dcn_mesh_shape. For example, an ici_mesh_shape of [2, 3, 4] with
		    mesh_axis_names ['replica', 'data', 'mdl'] indicates 2-way replica
		    parallelism, 3-way data parallelism, and 4-way model parallelism over 24
		    devices. None, the default, is equivalent to a sequence of ones and
		    means that the model is placed on a single device.
		  dcn_mesh_shape: Shape of the logical mesh used for SPMD parallelism over
		    multiple slices. The overall mesh is the product of ici_mesh_shape and
		    dcn_mesh_shape, and the meaning of each mesh axis is defined by
		    mesh_axis_names, so these three params must be the same length.
		"""

		if not num_partitions and not model_parallel_submesh:
			raise ValueError(
				"At least one of `num_partitions` or " "`model_parallel_submesh` must be set."
			)

		if model_parallel_submesh is not None and len(model_parallel_submesh) != 4:
			logging.error(
				(
					"`model_parallel_submesh` must be either None or a 4-tuple. Got"
					" `model_parallel_submesh`=%r. A ValueError will be raised"
					" beginning March 1, 2022."
				),
				model_parallel_submesh,
			)

		if bool(num_partitions) and bool(model_parallel_submesh):
			logging.error(
				"At most one of `num_partitions` or `model_parallel_submesh` can be "
				"set. Got `num_partitions=%r` and `model_parallel_submesh`=%r. A "
				"ValueError will be raised beginning March 21, 2022.",
				num_partitions,
				model_parallel_submesh,
			)

		self._num_partitions = num_partitions
		self._model_parallel_submesh = model_parallel_submesh
		self._params_on_devices = params_on_devices
		if ici_mesh_shape is None or dcn_mesh_shape is None:
			self._data_axis = "data"
		else:
			self._data_axis = ("replica", "data")
		self._backend = backend
		self._ici_mesh_shape = ici_mesh_shape
		self._dcn_mesh_shape = dcn_mesh_shape

	@property
	def mesh(self) -> Mesh:
		raise NotImplementedError

	@property
	def data_partition_spec(self) -> PartitionSpec:
		return PartitionSpec(self._data_axis)

	@property
	def data_mesh_size(self) -> int:
		"""Data mesh size.

		Data mesh size is defined as the number of global devices involved to
		carry out data parallel. Let's say we have a global mesh: ('replica': 2,
		'data': 4, 'model': 2), and axes 'replica' and 'data' are responsible for
		the data parallel, that means we have 2*4 = 8 devices involved - i.e., data
		mesh size is 8.

		Returns:
		  the id of the shard for the axes being replicated among the devices used
		  to shard the sharded_mesh_axes.
		"""
		data_submesh_sizes = (
			[self.mesh.shape[self._data_axis]]
			if isinstance(self._data_axis, str)
			else [self.mesh.shape[axis] for axis in self._data_axis]
		)
		data_mesh_size = functools.reduce(lambda x, y: x * y, data_submesh_sizes)
		return data_mesh_size

	@property
	def data_shards(self) -> int:
		"""Number of data shards.

		Let's say we are dealing with 2 slices of df4x2 TPUs. In data pipeline
		we need prepare / send one data shard to each local host. This means, we
		need 4 shards since we have 4 local hosts. How to infer the number of hosts
		from mesh information? In this case, we have a global mesh: ('replica': 2,
		'data': 8, 'model': 2). Each local host (i.e., df2x2) has this local mesh:
		('replica': 1, 'data': 4, 'model': 2). By dividing global mesh with local
		mesh, we can get the count of hosts.

		Returns:
		  Number of data shards. Each shard will be sent to one local host.
		"""
		data_chunks = (
			[self._local_chunker.num_chunks[self._data_axis]]
			if isinstance(self._data_axis, str)
			else [self._local_chunker.num_chunks[axis] for axis in self._data_axis]
		)
		data_shards = functools.reduce(lambda x, y: x * y, data_chunks)
		return data_shards

	@property
	def data_shard_id(self) -> int:
		"""Data shard id for the current host.

		Returns:
		  Index of data shard that will be sent to the current local host.
		"""
		return self._local_chunker.get_shard_id(self._data_axis)

	def get_data_layout(
		self, batch_size: Optional[int] = None, host_index: Optional[int] = None
	) -> DataLayout:
		"""Returns filled `DataLayout` based on the partitioned model layout.

		Args:
		  batch_size: if set, indicates the requested batch size. The exception will
		    be raised if this batch size is not compatible with the layout. If not
		    set, the batch size is inferred from the layout.
		  host_index: indicates the host index to use for the calculations, if not
		    set - use JAX-provided one. Should be in [0, num_hosts) interval and the
		    order should match the order of corresponding CPU devices in
		    `jax.devices()`.

		Returns:
		  Filled `DataLayout` structure.
		"""
		if host_index is not None:
			raise NotImplementedError("Explicit host_index is not yet implemented.")
		if self._data_axis is None:
			return DataLayout(
				batch_size=batch_size,
				shard_id=0,
				num_shards=1,
				is_first_host_in_replica_set=(jax.process_index() == 0),
			)

		batch_size = batch_size or self.data_mesh_size
		if batch_size % self.data_mesh_size:
			raise ValueError(
				f"Batch size ({batch_size}) must be divisible by corresponding "
				f"data mesh size ({self.data_mesh_size})."
			)

		if batch_size % self.data_shards:
			raise ValueError(
				f"Batch size ({batch_size}) must be divisible by number of "
				f"data shards ({self.data_shards})."
			)
		replica_id = self._local_chunker.get_replica_id(self._data_axis)
		return DataLayout(
			batch_size=int(batch_size),
			shard_id=int(self.data_shard_id),
			num_shards=int(self.data_shards),
			is_first_host_in_replica_set=(replica_id == 0),
		)

	def get_local_chunk_info(
		self, global_shape: Tuple[int, ...], mesh_axes: Sequence[Optional[str]]
	) -> LocalChunkInfo:
		"""Returns the local chunk info for a given array shape and sharded axes."""
		return self._local_chunker.get_local_chunk_info(global_shape, mesh_axes)

	@property
	def params_on_devices(self):
		return self._params_on_devices

	@params_on_devices.setter
	def params_on_devices(self, value):
		self._params_on_devices = value

	def move_params_to_devices(self, train_state, train_state_axes):
		"""Moves the optimizer parameters to devices."""
		p_id_fn = self.partition(
			_id_fn,
			in_axis_resources=(train_state_axes, None),
			out_axis_resources=(train_state_axes, None),
			donate_argnums=(0,),
		)
		if jax.process_count() > 1:
			train_state = host_local_array_to_global_array(
				train_state, self.mesh, train_state_axes
			)
		train_state, _ = p_id_fn(train_state, jnp.ones((), dtype=jnp.uint32))
		return train_state

	@property
	@abc.abstractmethod
	def _local_chunker(self):
		"""Returns the chunker that matches the parameters of this partitioner."""
		raise NotImplementedError

	def get_logical_axes(self, train_state):
		"""Returns a copy of TrainState with Optional[AxisNames] as leaves."""
		# By default, return None for the logical axes.
		return train_state.restore_state(
			jax.tree_util.tree_map(lambda x: None, train_state.state_dict())
		)

	def get_mesh_axes(self, train_state):
		"""Returns a copy of TrainState with Optional[PartitionSpecs] as leaves."""
		raise NotImplementedError

	@abc.abstractmethod
	def partition(
		self,
		fn: Callable,  # pylint: disable=g-bare-generic
		in_axis_resources,
		out_axis_resources,
		static_argnums: Union[int, Sequence[int]] = (),
		donate_argnums: Union[int, Sequence[int]] = (),
	) -> PartitionedCallable:
		raise NotImplementedError

	@abc.abstractmethod
	def compile(
		self, partitioned_fn: PartitionedCallable, *args
	) -> CompiledPartitionedCallable:
		raise NotImplementedError


class PjittedFnWithContext(PartitionedCallable):
	"""Wraps pjitted function to apply the appropriate contexts."""

	def __init__(
		self,
		pjitted_fn,
		partition_mesh: Mesh,
		logical_axis_rules: flax_partitioning.LogicalRules = (),
	):
		self._pjitted_fn = pjitted_fn
		self._mesh = partition_mesh
		self._logical_axis_rules = logical_axis_rules

	def __call__(self, *args, **kwargs):
		with (
			Mesh(self._mesh.devices, self._mesh.axis_names),
			flax_partitioning.axis_rules(self._logical_axis_rules),
		):
			return self._pjitted_fn(*args, **kwargs)

	def lower(self, *args, **kwargs):
		with (
			Mesh(self._mesh.devices, self._mesh.axis_names),
			flax_partitioning.axis_rules(self._logical_axis_rules),
		):
			return self._pjitted_fn.lower(*args, **kwargs)


class BasePjitPartitioner(BasePartitioner):
	"""Partitioner that uses T5X version of jax.pjit."""

	@cached_property
	def _local_chunker(self) -> LocalChunker:
		return LocalChunker(self.mesh)

	@cached_property
	def mesh(self) -> Mesh:
		return default_mesh(
			self._num_partitions,
			self._model_parallel_submesh,
			self._backend,
			self._ici_mesh_shape,
			self._dcn_mesh_shape,
		)

	def partition(
		self,
		fn: Callable,  # pylint: disable=g-bare-generic
		in_axis_resources,
		out_axis_resources,
		static_argnums: Union[int, Sequence[int]] = (),
		donate_argnums: Union[int, Sequence[int]] = (),
	) -> PjittedFnWithContext:
		pjitted = pjit(
			fn,
			in_shardings=in_axis_resources,
			out_shardings=out_axis_resources,
			static_argnums=static_argnums,
			donate_argnums=donate_argnums,
		)

		return PjittedFnWithContext(pjitted, self.mesh)

	def compile(
		self, partitioned_fn: PjittedFnWithContext, *args
	) -> CompiledPartitionedCallable:
		return partitioned_fn.lower(*args).compile()


class PjitPartitioner(BasePjitPartitioner):
	"""Partitioner that uses named axes and jax.pjit."""

	def __init__(
		self,
		num_partitions: Optional[int] = None,
		model_parallel_submesh: Optional[HardwareMesh] = None,
		params_on_devices: bool = True,
		backend: Optional[str] = None,
		ici_mesh_shape: Optional[HardwareMesh] = None,
		dcn_mesh_shape: Optional[HardwareMesh] = None,
		logical_axis_rules: Optional[LogicalAxisRules] = None,
	):
		"""PjitPartitioner constructor.

		See https://github.com/google-research/text-to-text-transfer-transformer/blob/main/README.mdx/usage/partitioning for details.

		Args:
		  num_partitions: an integer that specifies the size of the model parallel
		    submesh to be automatically selected for the current topology. See
		    `model_parallel_submesh` for details on how this submesh is used.
		    Mutually exclusive with `model_parallel_submesh`.
		  model_parallel_submesh: is a 4-tuple that specifies the `(x, y, z, c)`
		    submesh model-parallel device tile, an axis of accelerator parallelism
		    orthogonal to data parallelism. Array axes in a model's parameters or
		    activations can be sharded over this submesh using axis rules (see
		    `logical_axis_rules`) that map them to 'model'. The effective number of
		    model sub-partitions is equal to `np.prod(model_parallel_submesh)` and
		    must evenly divide the total number of devices (i.e.,
		    `jax.device_count() % np.prod(model_parallel_submesh) == 0`). The rest
		    of the TPU mesh is the data parallel submesh, providing
		    `jax.device_count() // np.prod(model_parallel_submesh)` partitions. It
		    is used for data (batch) parallelism and to shard other array axes that
		    are mapped to 'data'. This argument is mutually exclusive with
		    `num_partitions`.
		  params_on_devices: whether to keep the params on devices, if False -
		    params stay in the host memory. Note that some partitioners might ignore
		    this setting, for example if they don't support storing all params on
		    device memory.
		  backend: get devices from the pinned backend, if specified. This is useful
		    for explicitly specifying the devices other than relying on
		    jax_platform_name.
		  ici_mesh_shape: Shape of the logical mesh used for SPMD parallelism in
		    each slice. The meaning of each mesh axis is defined by mesh_axis_names,
		    so these two params must be the same length. If dcn_mesh_shape is
		    present, the overall mesh is the product of ici_mesh_shape and
		    dcn_mesh_shape. For example, an ici_mesh_shape of [2, 3, 4] with
		    mesh_axis_names ['replica', 'data', 'model'] indicates 2-way replica
		    parallelism, 3-way data parallelism, and 4-way model parallelism over 24
		    devices. None, the default, is equivalent to a sequence of ones and
		    means that the model is placed on a single device.
		  dcn_mesh_shape: Shape of the logical mesh used for SPMD parallelism over
		    multiple slices. The overall mesh is the product of ici_mesh_shape and
		    dcn_mesh_shape, and the meaning of each mesh axis is defined by
		    mesh_axis_names, so these three params must be the same length.
		  logical_axis_rules: a priority-ordered sequence of KV tuples that maps
		    logical axis names to either `None` (not sharded), 'model' (to shard
		    across the model-parallel submesh), or 'data' (to shard across the
		    data-parallel submesh).
		"""
		super().__init__(
			num_partitions=num_partitions,
			model_parallel_submesh=model_parallel_submesh,
			params_on_devices=params_on_devices,
			backend=backend,
			ici_mesh_shape=ici_mesh_shape,
			dcn_mesh_shape=dcn_mesh_shape,
		)
		if logical_axis_rules is None:
			logical_axis_rules = standard_logical_axis_rules()
		if ici_mesh_shape is not None and dcn_mesh_shape is not None:
			# Split batch over new replica axis.
			logical_axis_rules = (
				(k, ("replica", "data") if k == "batch" else v) for k, v in logical_axis_rules
			)
		self._logical_axis_rules = tuple(logical_axis_rules)
		(self._data_axis,) = flax_partitioning.logical_to_mesh_axes(
			["batch"], self._logical_axis_rules
		)

	def partition(
		self,
		fn: Callable,  # pylint: disable=g-bare-generic
		in_axis_resources,
		out_axis_resources,
		static_argnums: Union[int, Sequence[int]] = (),
		donate_argnums: Union[int, Sequence[int]] = (),
	) -> PjittedFnWithContext:
		"""Partitions the function using jax.pjit."""
		pjitted = pjit(
			fn,
			in_shardings=in_axis_resources,
			out_shardings=out_axis_resources,
			static_argnums=static_argnums,
			donate_argnums=donate_argnums,
		)

		return PjittedFnWithContext(pjitted, self.mesh, self._logical_axis_rules)

	@property
	def logical_axis_rules(self):
		"""Returns the logical axis rules."""
		return self._logical_axis_rules

	def get_logical_axes(self, train_state):
		"""Returns a copy of TrainState with Optional[AxisNames] as leaves."""
		return train_state.as_logical_axes()

	def get_mesh_axes(self, train_state):
		"""Returns a copy of TrainState with Optional[PartitionSpecs] as leaves."""
		logical_axes = self.get_logical_axes(train_state)

		def _logical_to_mesh_axes(param_name, logical_axes):
			if logical_axes is None:
				return None
			elif logical_axes is traverse_util.empty_node:
				return traverse_util.empty_node
			try:
				return flax_partitioning.logical_to_mesh_axes(
					logical_axes, self._logical_axis_rules
				)
			except ValueError as e:
				raise ValueError(f"Failed to map logical axes for {param_name}") from e

		flat_logical_axes = traverse_util.flatten_dict(
			logical_axes.state_dict(), keep_empty_nodes=True, sep="/"
		)
		flat_mesh_axes = {
			k: _logical_to_mesh_axes(k, v) for k, v in flat_logical_axes.items()
		}

		return logical_axes.restore_state(
			traverse_util.unflatten_dict(flat_mesh_axes, sep="/")
		)


def host_local_array_to_global_array(arr_tree, mesh: jax.sharding.Mesh, pspecs):
	pspecs = jax.tree_util.tree_map(
		lambda x: PartitionSpec() if x is None else x,
		pspecs,
		is_leaf=lambda x: x is None,
	)
	return multihost_utils.host_local_array_to_global_array(arr_tree, mesh, pspecs)



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sparsity distribution module with functions generates sparsity tree.

For Jax Pruner, users can allocate different desired sparsity for each different
parameter (weight), so the functions in this file help users to allocate the
sparsity to meet the overall target sparsity.

Note that all functions in this file must have the same input/output signatures.
"""

from typing import NamedTuple, Tuple, Union


class Unstructured(NamedTuple):
	pass


class Block(NamedTuple):
	# This block is allocated for the last two dims of the corresponding variable.
	block_shape: Tuple[int, int]
	# If true, pool type is AVG. Otherwise, pool type is MAX.
	use_avg_pooling: bool = True


class NByM(NamedTuple):
	# This n-by-m operation works for the target axis of the corresponding
	# variable.
	n: int
	m: int
	axis: int = -1


class Channel(NamedTuple):
	# This channel-wise operations works for the target axis of the corresponding
	# variable.
	axis: int = -1


SparsityType = Union[Unstructured, Block, NByM, Channel]


# TODO following is needed since isintance(obj, SparsityType) fails.
# TODO check whether there is a better way.
def is_sparsity_type(instance_to_check):
	"""Checks whether given object is a sparsity type."""
	return type(instance_to_check) in [Unstructured, Block, NByM, Channel]



from fjformer.optimizers.adamw import (
	get_adamw_with_cosine_scheduler as get_adamw_with_cosine_scheduler,
	get_adamw_with_warmup_cosine_scheduler as get_adamw_with_warmup_cosine_scheduler,
	get_adamw_with_warmup_linear_scheduler as get_adamw_with_warmup_linear_scheduler,
	get_adamw_with_linear_scheduler as get_adamw_with_linear_scheduler,
)
from fjformer.optimizers.lion import (
	get_lion_with_cosine_scheduler as get_lion_with_cosine_scheduler,
	get_lion_with_warmup_linear_scheduler as get_lion_with_warmup_linear_scheduler,
	get_lion_with_warmup_cosine_scheduler as get_lion_with_warmup_cosine_scheduler,
	get_lion_with_linear_scheduler as get_lion_with_linear_scheduler,
)
from fjformer.optimizers.adafactor import (
	get_adafactor_with_cosine_scheduler as get_adafactor_with_cosine_scheduler,
	get_adafactor_with_warmup_cosine_scheduler as get_adafactor_with_warmup_cosine_scheduler,
	get_adafactor_with_warmup_linear_scheduler as get_adafactor_with_warmup_linear_scheduler,
	get_adafactor_with_linear_scheduler as get_adafactor_with_linear_scheduler,
)

from fjformer.optimizers.rmsprop import (
	get_rmsprop_with_cosine_scheduler as get_rmsprop_with_cosine_scheduler,
	get_rmsprop_with_linear_scheduler as get_rmsprop_with_linear_scheduler,
	get_rmsprop_with_warmup_linear_scheduler as get_rmsprop_with_warmup_linear_scheduler,
	get_rmsprop_with_warmup_cosine_scheduler as get_rmsprop_with_warmup_cosine_scheduler,
)

from fjformer.optimizers.optimizer_utils import (
	optax_add_scheduled_weight_decay as optax_add_scheduled_weight_decay,
)



from typing import Any, Optional, Tuple, Type, TypeVar

from plum import dispatch, parametric


# Custom exceptions
class ComplementError(Exception):
	"""Base exception class for Complement-related errors."""

	pass


class TypeParameterError(ComplementError):
	"""Exception raised for errors in type parameter initialization or comparison."""

	pass


# Type variables for better type hinting
A = TypeVar("A")
B = TypeVar("B")


class _ComplementMeta(type):
	def __instancecheck__(self, x: Any) -> bool:
		"""
		Check if an object is an instance of the Complement type.

		Args:
		    x: The object to check.

		Returns:
		    bool: True if x is an instance of Complement, False otherwise.
		"""
		try:
			a, b = self.type_parameter
			return a is None or (isinstance(x, a) and not isinstance(x, b))
		except Exception as e:
			raise TypeParameterError("Error in instance check") from e


@parametric
class Complement(metaclass=_ComplementMeta):
	"""
	Represents the relative complement of two types.

	The Complement[A, B] represents all elements in A that are not in B (A - B).

	Attributes:
	    type_parameter (Tuple[Optional[Type], Optional[Type]]): A tuple containing the two types A and B.
	"""

	@classmethod
	@dispatch
	def __init_type_parameter__(
		cls,
		a: Optional[Type[A]],
		b: Optional[Type[B]],
	) -> Tuple[Optional[Type[A]], Optional[Type[B]]]:
		"""
		Initialize the type parameters for the Complement.

		Args:
		    a: The first type (superset).
		    b: The second type (subset to be removed from a).

		Returns:
		    A tuple containing the two type parameters.

		Raises:
		    TypeParameterError: If there's an error initializing the type parameters.
		"""
		try:
			return a, b
		except Exception as e:
			raise TypeParameterError("Error initializing type parameters") from e

	@classmethod
	@dispatch
	def __le_type_parameter__(
		cls,
		left: Tuple[Optional[Type[A]], Optional[Type[B]]],
		right: Tuple[Optional[Type[A]], Optional[Type[B]]],
	) -> bool:
		"""
		Compare two Complement type parameters for the less than or equal relationship.

		This method defines the subtyping relationship between two Complement types.

		Args:
		    left: The left-hand side Complement type parameters.
		    right: The right-hand side Complement type parameters.

		Returns:
		    bool: True if the left Complement is a subtype of the right Complement, False otherwise.

		Raises:
		    TypeParameterError: If there's an error comparing the type parameters.
		"""
		try:
			a_left, b_left = left
			a_right, b_right = right

			# Check if a_left is a subclass of a_right and b_right is a subclass of b_left
			is_subclass_a = issubclass(a_left, a_right) if a_left and a_right else True
			is_subclass_b = issubclass(b_right, b_left) if b_right and b_left else True

			return is_subclass_a and is_subclass_b
		except Exception as e:
			raise TypeParameterError("Error comparing type parameters") from e

	@classmethod
	def is_complement(cls, x: Any) -> bool:
		"""
		Check if an object is an instance of any Complement type.

		Args:
		    x: The object to check.

		Returns:
		    bool: True if x is an instance of any Complement type, False otherwise.
		"""
		return isinstance(x, Complement)



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the Sparsity updater classes in base_updater.py."""

import functools
from unittest import mock

import chex
import jax.numpy as jnp
from absl.testing import absltest, parameterized

from fjformer.jaxpruner import (
	base_updater,
	sparsity_distributions,
	sparsity_schedules,
	sparsity_types,
)


class BaseUpdaterTest(parameterized.TestCase, absltest.TestCase):
	def setUp(self):
		super(BaseUpdaterTest, self).setUp()
		self.updater = base_updater.BaseUpdater(
			scheduler=sparsity_schedules.PolynomialSchedule(
				update_freq=2, update_start_step=4, update_end_step=9
			),
			sparsity_distribution_fn=functools.partial(
				sparsity_distributions.uniform, sparsity=0.8
			),
		)

	def testUpdaterInitialized(self):
		self.assertIsInstance(self.updater.sparsity_type, sparsity_types.Unstructured)
		self.assertTrue(callable(self.updater.topk_fn))

	def testOptaxWrapper(self):
		distribution_fn = mock.MagicMock()
		sparse_updater = base_updater.BaseUpdater(
			scheduler=sparsity_schedules.PolynomialSchedule(
				update_freq=1, update_start_step=0, update_end_step=1
			),
			sparsity_distribution_fn=distribution_fn,
		)
		sparse_updater.get_initial_masks = mock.MagicMock()
		sparse_updater.maybe_update_mask = mock.MagicMock()
		sparse_updater.maybe_update_mask.return_value = ({}, {})

		inner = mock.MagicMock()
		inner.update.return_value = ({}, None)

		tx = sparse_updater.wrap_optax(inner)

		_ = tx.init({})
		distribution_fn.assert_called_once()
		sparse_updater.get_initial_masks.assert_called_once()

		sparse_updater.calculate_scores = mock.MagicMock()
		sparse_updater.calculate_scores.return_value = None

		_ = tx.update({}, base_updater.SparseState(masks=None, count=0), {})
		inner.update.assert_called_once()
		sparse_updater.calculate_scores.assert_called_once()

		# Test passing an extra argument to update().
		inner = mock.MagicMock()
		inner.update.return_value = ({}, None)
		sparse_updater.calculate_scores.reset_mock()
		inner.update.assert_not_called()

		tx = sparse_updater.wrap_optax(inner)
		extra_arg = object()
		_ = tx.update({}, base_updater.SparseState(masks=None, count=0), {}, foo=extra_arg)
		inner.update.assert_called_once()
		_, kwargs = inner.update.call_args
		self.assertEmpty(kwargs, "Should not pass extra args to inner")
		sparse_updater.calculate_scores.assert_called_once()
		_, kwargs = sparse_updater.calculate_scores.call_args
		self.assertDictContainsSubset({"foo": extra_arg}, kwargs)

	def testInstantSparsify(self):
		distribution_fn = mock.MagicMock()
		distribution_fn.return_value = {}
		sparse_updater = base_updater.BaseUpdater(
			scheduler=sparsity_schedules.PolynomialSchedule(
				update_freq=1, update_start_step=0, update_end_step=1
			),
			sparsity_distribution_fn=distribution_fn,
		)
		sparse_updater.calculate_scores = mock.MagicMock()
		sparse_updater.calculate_scores.return_value = {}
		sparse_updater.create_masks = mock.MagicMock()
		masks_returned = {"a": jnp.array([1, 0])}
		sparse_updater.create_masks.return_value = masks_returned
		masked_vars, masks = sparse_updater.instant_sparsify({"a": jnp.array([2.0, -3.0])})
		chex.assert_trees_all_close(masks, masks_returned)
		chex.assert_trees_all_close(masked_vars, {"a": jnp.array([2.0, 0.0])})
		distribution_fn.assert_called_once()
		sparse_updater.calculate_scores.assert_called_once()

		# Test passing an extra argument to instant_sparsify().
		sparse_updater.calculate_scores.reset_mock()
		extra_arg = object()
		_ = sparse_updater.instant_sparsify({"a": jnp.array([2.0, -3.0])}, foo=extra_arg)
		sparse_updater.calculate_scores.assert_called_once()
		_, kwargs = sparse_updater.calculate_scores.call_args
		self.assertDictContainsSubset({"foo": extra_arg}, kwargs)

	def testNoPruning(self):
		updater = base_updater.NoPruning()
		param_tree = {"a": 0, "b": {"c": 0, "d": 0}}
		expected = {"a": None, "b": {"c": None, "d": None}}
		self.assertEqual(
			updater.get_initial_masks(param_tree, target_sparsities=None), expected
		)
		chex.assert_trees_all_close(updater.instant_sparsify(param_tree)[0], param_tree)

	def testCustomTopkFn(self):
		def _mock_topk_fn(t, _):
			return jnp.ones_like(t)

		sparse_updater = base_updater.BaseUpdater(
			topk_fn=_mock_topk_fn,
		)
		# Test if custom topk_fn is not overrided.
		self.assertEqual(sparse_updater.topk_fn, _mock_topk_fn)


if __name__ == "__main__":
	absltest.main()



import os

import msgpack
from flax.serialization import from_bytes, to_bytes, to_state_dict
from flax.traverse_util import flatten_dict
from jax import numpy as jnp
from typing import Union

from fjformer.checkpoint.streamer import CheckpointManager


def get_dtype(dtype):
	if isinstance(dtype, str):
		dtype = {
			"bf16": jnp.bfloat16,
			"bfloat16": jnp.bfloat16,
			"fp16": jnp.float16,
			"float16": jnp.float16,
			"fp32": jnp.float32,
			"float32": jnp.float32,
			"fp64": jnp.float64,
			"float64": jnp.float64,
		}[dtype]
	return dtype


def float_tensor_to_dtype(tensor, dtype):
	if dtype is None or dtype == "":
		return tensor
	if isinstance(dtype, str):
		dtype = get_dtype(dtype)
	float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)
	if getattr(tensor, "dtype", None) in float_dtypes:
		tensor = tensor.astype(dtype)
	return tensor


def load_and_convert_checkpoint_to_torch(
	path,
	dtype=jnp.float16,
	transpose_needed=None,
	transpose_not_needed=None,
	select_params_field: bool = True,
):
	import torch

	if transpose_needed is None:
		transpose_needed = ["kernel"]
	if transpose_not_needed is None:
		transpose_not_needed = ["none"]

	def match_keywords(string, ts, ns):
		for t in ts:
			if t not in string:
				return False
		for n in ns:
			if n in string:
				return False
		return True

	_, flax_params = CheckpointManager.load_state_checkpoint("params", path)
	flax_params = (
		flatten_dict(flax_params["params"], sep=".")
		if select_params_field
		else flatten_dict(flax_params, sep=".")
	)
	torch_params = {}
	for key, tensor in flax_params.items():
		if match_keywords(key, transpose_needed, transpose_not_needed):
			tensor = tensor.T
		tensor = float_tensor_to_dtype(tensor, dtype)
		torch_params[key] = torch.from_numpy(tensor)
	return torch_params


def read_ckpt(
	path: Union[str, os.PathLike], shard_fns=None, add_extra_past_fix: list = None
):
	tensors = {}
	with open(path, "rb") as stream:
		unpacker = msgpack.Unpacker(stream, read_size=83886080, max_buffer_size=0)
		for key, value in unpacker:
			if add_extra_past_fix is not None:
				key = add_extra_past_fix + key
			key = tuple(key)
			tensor = from_bytes(None, value)
			if shard_fns is not None:
				tensor = shard_fns[key](tensor)
			tensors[key] = tensor
	return tensors


def save_ckpt(train_state, path, gather_fns=None, float_dtype=None):
	train_state = to_state_dict(train_state)
	packer = msgpack.Packer()
	flatten_train_state = flatten_dict(train_state)
	if gather_fns is not None:
		gather_fns = flatten_dict(to_state_dict(gather_fns))

	with open(path, "wb") as stream:
		for key, value in flatten_train_state.items():
			if gather_fns is not None:
				value = gather_fns[key](value)
			value = float_tensor_to_dtype(value, float_dtype)
			stream.write(packer.pack((key, to_bytes(value))))



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file implements common global pruning algorithms."""

import dataclasses
from typing import Dict, Optional

import flax
import jax
import jax.numpy as jnp

from fjformer.jaxpruner import base_updater, sparsity_distributions, sparsity_types
from fjformer.jaxpruner.algorithms import pruners


@dataclasses.dataclass
class GlobalPruningMixin:
	"""Implements Mixin for global pruning."""

	custom_sparsity_map: Optional[Dict[str, float]] = None
	filter_fn: base_updater.FilterFnType = sparsity_distributions.NOT_DIM_ONE_FILTER_FN
	normalization_eps: float = 1e-8
	use_normalization: bool = True
	sparsity: float = 0.0

	def init_state(self, params):
		"""Regular init. but target_sparsities are set to orignal sparsity."""
		if not isinstance(self.sparsity_type, sparsity_types.Unstructured):
			raise AttributeError(
				f"Sparsity type {self.sparsity_type.__class__} is not supported for"
				" GlobalPruningMixin."
			)
		masks = self.create_masks(params, 0.0)
		if self.use_packed_masks:
			masks = jax.tree_map(jnp.packbits, masks)
		# Global sparsity only needs global target sparsity.
		return base_updater.SparseState(
			masks=masks,
			target_sparsities=self.sparsity,
			count=jnp.zeros([], jnp.int32),
		)

	def instant_sparsify(self, params, grads=None):
		# Global sparsity doesn't require sparsity distribution as it uses
		# global ordering of scores.
		scores = self.calculate_scores(params, grads=grads)
		masks = self.create_masks(scores, self.sparsity)
		if self.use_packed_masks:
			masks = jax.tree_map(jnp.packbits, masks)
		return self.apply_masks(params, masks, is_packed=False), masks

	def create_masks(self, scores, target_sparsity):
		"""Creates masks using global ordering."""
		custom_sparsity_map = self.custom_sparsity_map or {}

		def _unified_filter_fn(k, score):
			return self.filter_fn(k, score) and k not in custom_sparsity_map

		def _maybe_normalize(score):
			if self.use_normalization:
				return score / (jnp.linalg.norm(score) + self.normalization_eps)
			else:
				return score

		flat_scores = flax.traverse_util.flatten_dict(scores)
		filtered_scores = {
			k: _maybe_normalize(score)
			for k, score in flat_scores.items()
			if _unified_filter_fn(k, score)
		}
		ordered_keys = sorted(filtered_scores.keys())

		filtered_scores_concat = jnp.concatenate(
			[filtered_scores[k] for k in ordered_keys], axis=None
		)

		flat_mask_concat = self.topk_fn(filtered_scores_concat, target_sparsity)
		res_dict = {}
		cur_index = 0
		for k in ordered_keys:
			param = filtered_scores[k]
			next_index = cur_index + param.size
			flat_mask = flat_mask_concat[cur_index:next_index]
			res_dict[k] = jnp.reshape(flat_mask, param.shape)
			cur_index = next_index

		for k, score in flat_scores.items():
			if k in ordered_keys:
				pass
			elif k in custom_sparsity_map:
				res_dict[k] = self.topk_fn(score, custom_sparsity_map[k])
			else:
				res_dict[k] = None
		return_val = flax.traverse_util.unflatten_dict(res_dict)
		if isinstance(scores, flax.core.frozen_dict.FrozenDict):
			return_val = flax.core.freeze(return_val)
		return return_val


@dataclasses.dataclass
class GlobalMagnitudePruning(GlobalPruningMixin, pruners.MagnitudePruning):
	"""Magnitude pruner with global ordering."""

	pass


@dataclasses.dataclass
class GlobalSaliencyPruning(GlobalPruningMixin, pruners.SaliencyPruning):
	"""Saliency pruner with global ordering."""

	pass



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Union

import flax.core
import flax.traverse_util
import jax
import jax.tree_util
import optax
from jax import numpy as jnp

from fjformer.core import (
	EmptyNode,
	implicit_compact,
	materialize_nested,
	tree_map_with_implicit,
)
from fjformer.core.utilities import freeze_keys, freeze_subtrees
from fjformer.lora.lora_core import LoraWeight
from fjformer.utils import get_logger

LORA_FREEZE = 0
LORA_FULL = -1

logger = get_logger(__name__)


def split_lora_params(
	params: Union[dict, jax.tree_util.PyTreeDef],
	lora_spec: Union[dict, jax.tree_util.PyTreeDef],
) -> Union[dict, jax.tree_util.PyTreeDef]:
	"""Replaces specific parameters in a PyTree with `EmptyNode`.

	This function processes a PyTree of parameters, replacing `LoraWeight.w`
	values and parameters marked with `LORA_FREEZE` in the `lora_spec` with
	`EmptyNode` instances. This is particularly useful for checkpointing,
	allowing the saving of only trainable parameters.

	Args:
	    params: A PyTree of model parameters.
	    lora_spec: A PyTree mirroring the structure of `params`, with values
	        indicating how each parameter should be handled (LORA_FREEZE,
	        LORA_FULL, or a LoRA dimension).

	Returns:
	    A new PyTree with the specified parameters replaced by `EmptyNode`.
	"""

	def node_mapper(node: Any, spec_val: Any) -> Any:
		if not isinstance(node, LoraWeight):
			return node if spec_val != LORA_FREEZE else EmptyNode
		children, aux = node.tree_flatten_with_keys()
		idx = next(i for i, (key, _) in enumerate(children) if key == "w")
		children[idx] = ("w", EmptyNode)
		return LoraWeight.tree_unflatten(aux, [c for _, c in children])

	return tree_map_with_implicit(node_mapper, params, lora_spec)


@dataclass
class RaptureConfig:
	"""Configuration for the RaptureConfig fine-tuning method.

	Attributes:
	    lora_dim: Dimensionality of LoRA adapters.
	    fully_fine_tune_parameters: Optional list of parameter names to be fully fine-tuned.
	    lora_fine_tune_parameters: Optional list of parameter names to be fine-tuned with LoRA.
	    tune_vectors: Whether to tune vector parameters (length 1).
	    verbose: If True, logs information about parameter tuning decisions.
	    dtype: The dtype of the LoRA parameters.
	"""

	lora_dim: int
	fully_fine_tune_parameters: Optional[List[str]] = None
	lora_fine_tune_parameters: Optional[List[str]] = None
	tune_vectors: bool = True
	verbose: bool = True
	dtype: jnp.dtype = jnp.float32


@dataclass
class RaptureModule:
	"""Container for data related to a Flax module adapted with RaptureConfig.

	Attributes:
	    lora_opt_state: The optimizer state for the LoRA parameters.
	    lora_parameters: The PyTree of LoRA parameters.
	    lora_module: The original Flax module adapted for use with LoRA parameters.
	    lora_specs: A PyTree mirroring the parameter structure, indicating the tuning strategy for each parameter.
	    lora_tx: The Optax gradient transformation wrapped for LoRA.
	"""

	lora_opt_state: Union[jax.tree_util.PyTreeDef, dict]
	lora_parameters: Union[jax.tree_util.PyTreeDef, dict]
	lora_module: Union[Any, flax.linen.Module]
	lora_specs: Union[jax.tree_util.PyTreeDef, dict]
	lora_tx: optax.GradientTransformation


class LoraRapture:
	"""Implements the RaptureConfig fine-tuning method.

	RaptureConfig combines LoRA (Low-Rank Adaptation) with full fine-tuning,
	allowing for efficient adaptation of large language models.
	"""

	def __init__(self, config: RaptureConfig):
		"""Initializes RaptureConfig with the provided configuration.

		Args:
		    config: An `RaptureConfig` object specifying the fine-tuning setup.
		"""
		self.config = config

	@staticmethod
	def merge_parameters(
		lora_parameters: Union[dict, jax.tree_util.PyTreeDef],
		destructive: bool = False,
	) -> Union[dict, jax.tree_util.PyTreeDef]:
		"""Merges LoRA parameters into the base model parameters.

		This function iterates through the PyTree of LoRA parameters, merging
		them with the corresponding base parameters within `LoraWeight`
		instances.  Optionally, it can destructively modify the input
		`lora_parameters` to free memory.

		Args:
		    lora_parameters: A PyTree of LoRA parameters.
		    destructive: If True, the input `lora_parameters` may be
		        destructively modified to free memory.

		Returns:
		    A new PyTree with LoRA parameters merged into the base parameters.
		"""

		def _ensure_delete(val: jnp.ndarray) -> None:
			if not isinstance(val, jax.Array) or val.is_deleted():
				return
			try:
				val.device_buffer.delete()  # type: ignore
			except (ValueError, AttributeError):
				val.addressable_data(0).delete()

		materialize = jax.jit(materialize_nested, donate_argnums=0 if destructive else ())

		def map_fn(param: Any) -> Any:
			if isinstance(param, LoraWeight):
				result = materialize(param)
				if destructive:
					jax.tree_util.tree_map(_ensure_delete, param)
				return result
			return param

		return tree_map_with_implicit(map_fn, lora_parameters)

	def get_lora_parameters(self, lora_parameters) -> Dict:
		"""
		Extract LoRA (Low-Rank Adaptation) parameters from a parameter tree.

		This method traverses the given parameter tree and extracts LoRA-specific
		parameters (a, b, and alpha) from LoraWeight instances. It flattens the
		resulting structure, removes any None values, and then unflattens it back
		into a nested dictionary.

		Args:
		    lora_parameters (Any): A nested structure containing model parameters,
		                        potentially including LoraWeight instances.

		Returns:
		    Dict: A nested dictionary containing only the LoRA parameters. Each LoRA
		        parameter set is represented as a dict with keys 'a', 'b', and 'alpha'.

		Note:
		    - This method uses `tree_map_with_implicit` to traverse the parameter tree.
		    - Parameters that are not instances of LoraWeight are mapped to None
		    and subsequently removed from the result.
		"""

		def map_fn(param: Any) -> Dict[str, Any] | None:
			"""
			Extract LoRA parameters from a single parameter.

			Args:
			    param (Any): A single parameter from the parameter tree.

			Returns:
			    Dict[str, Any] | None: A dictionary containing LoRA parameters
			                        if the input is a LoraWeight instance,
			                        otherwise None.
			"""
			if isinstance(param, LoraWeight):
				return {"a": param.a, "b": param.b, "alpha": param.alpha}
			return None

		pure_map = tree_map_with_implicit(map_fn, lora_parameters)
		pure_map = flax.traverse_util.flatten_dict(pure_map)
		keys = list(pure_map.keys())
		for key in keys:
			if pure_map[key] is None:
				pure_map.pop(key)
		return flax.traverse_util.unflatten_dict(pure_map)

	def base_decision_function(
		self,
		path: List[jax.tree_util.DictKey],
		params: Optional[Union[dict, jax.tree_util.PyTreeDef]] = None,
	) -> int:
		"""Determines the fine-tuning strategy for a parameter based on its path.

		This function acts as the default decision function for RaptureConfig. It
		checks if a parameter's path matches any entry in
		`fully_fine_tune_parameters` or `lora_fine_tune_parameters` from
		the `RaptureConfig`.  If a match is found, it returns `LORA_FULL`
		or the configured `lora_dim` respectively.  Otherwise, it defaults to
		`LORA_FREEZE`, indicating the parameter should not be fine-tuned.

		Args:
		    path: A list of `jax.tree_util.DictKey` representing the path to
		        the parameter within the model's parameter PyTree.
		    params: This argument is unused and can be omitted.

		Returns:
		    An integer representing the fine-tuning strategy for the parameter:
		        * `LORA_FULL`: Full fine-tuning.
		        * `LORA_FREEZE`: Freeze the parameter (no fine-tuning).
		        * `self.config.lora_dim`: Fine-tune using LoRA with the
		            configured dimensionality.
		"""
		del params  # Unused

		if self.config.fully_fine_tune_parameters is not None:
			for param_name in self.config.fully_fine_tune_parameters:
				if jax.tree_util.DictKey(key=param_name) in path:
					if self.config.verbose:
						logger.info(
							f"Array {'/'.join(str(n.key) for n in path)} "
							f"selected for full fine-tuning."
						)
					return LORA_FULL

		if self.config.lora_fine_tune_parameters is not None:
			for param_name in self.config.lora_fine_tune_parameters:
				if jax.tree_util.DictKey(key=param_name) in path:
					if self.config.verbose:
						logger.info(
							f"Array {'/'.join(str(n.key) for n in path)} "
							f"converted to LoraWeight with {self.config.lora_dim} dimensions."
						)
					return self.config.lora_dim

		return LORA_FREEZE

	def make_lora_specs(
		self,
		parameters: Union[dict, flax.core.FrozenDict],
		decision_fn: Optional[Callable] = None,
		tune_vectors: bool = False,
	) -> Union[dict, jax.tree_util.PyTreeDef]:
		"""Creates a PyTree specifying the fine-tuning strategy for parameters.

		This function generates a PyTree matching the structure of the input
		`parameters`.  Each value in the output tree represents the
		fine-tuning strategy for the corresponding parameter in `parameters`,
		determined by the provided `decision_fn`.

		Args:
		    parameters: A PyTree of model parameters.
		    decision_fn: A callable that takes the parameter path and the
		        parameter array as input and returns an integer representing the
		        fine-tuning strategy (`LORA_FREEZE`, `LORA_FULL`, or a
		        LoRA dimension).  If None, `self.base_decision_function` is used.
		    tune_vectors: Whether to tune vector parameters. If True, vectors
		        will be fully fine-tuned.

		Returns:
		    A PyTree mirroring the structure of `parameters`, with values
		    indicating the fine-tuning strategy for each parameter.
		"""
		decision_fn = (
			decision_fn if decision_fn is not None else self.base_decision_function
		)

		if decision_fn is None:

			def decision_fn(*args):  # pylint: disable=function-redefined
				return LORA_FREEZE

		def full_fn(path: List[jax.tree_util.DictKey], arr: jnp.ndarray) -> int:
			if len(arr.shape) < 2:
				return LORA_FULL if tune_vectors else LORA_FREEZE
			return decision_fn(path, arr)

		return jax.tree_util.tree_map_with_path(full_fn, parameters, is_leaf=None)

	@staticmethod
	def init_lora_parameters(
		param_tree: Union[dict, jax.tree_util.PyTreeDef],
		lora_spec: Union[dict, jax.tree_util.PyTreeDef],
		dtype: jnp.dtype = jnp.float32,
		rng: Optional[jax.random.PRNGKey] = None,
		stddev: float = 0.01,
		alpha: float = 1.0,
		is_leaf: bool = None,
	) -> Union[dict, jax.tree_util.PyTreeDef]:
		"""Initializes LoRA parameters based on the provided specifications.

		This function traverses the `param_tree`, initializing LoRA parameters
		according to the `lora_spec`.  Parameters marked for full
		fine-tuning (`LORA_FULL`) or freezing (`LORA_FREEZE`) retain their
		original form.  Parameters designated for LoRA tuning are replaced
		with `LoraWeight` instances, with adapter matrices initialized
		using a normal distribution.

		Args:
		    param_tree: A PyTree of model parameters.
		    lora_spec: A PyTree mirroring the structure of `param_tree`,
		        specifying the fine-tuning strategy for each parameter.
		    dtype: The data type for the LoRA parameters.
		    rng: A JAX PRNG key for random initialization.
		    stddev: Standard deviation for the normal distribution used to
		        initialize LoRA adapter matrices.
		    alpha: An additional scaling factor for the LoRA adapters.
		    is_leaf: Optional function to determine if a node in the PyTree
		        is a leaf.

		Returns:
		    A PyTree matching the structure of `param_tree` with LoRA
		    parameters initialized according to `lora_spec`.

		Raises:
		    ValueError: If a vector parameter is encountered with a
		        LoRA dimension specification (neither frozen nor fully tuned).
		"""
		if rng is None:
			rng = jax.random.PRNGKey(0)

		def iter_keys(key: jax.random.PRNGKey) -> jax.random.PRNGKey:
			while True:
				key, out_key = jax.random.split(key)
				yield out_key

		key_it = iter_keys(rng)

		def get_param(
			path: List[jax.tree_util.DictKey], param: jnp.ndarray, spec_val: int
		) -> Any:
			if spec_val in (LORA_FREEZE, LORA_FULL):
				return param

			if len(param.shape) == 1:
				raise ValueError(
					f"Vectors must either be frozen or fully tuned, but got "
					f"lora_spec value {lora_spec} for param with path {path}"
				)

			if len(param.shape) == 2:
				b_dim, a_dim = param.shape
				b = jnp.zeros((b_dim, spec_val), dtype=dtype)
				a = jax.random.normal(next(key_it), (spec_val, a_dim), dtype=dtype) * stddev
				return LoraWeight(w=param, a=a, b=b, alpha=alpha)

			# Convolutional parameter case
			*window_shape, in_channels, out_channels = param.shape
			a = jnp.zeros(
				(*(1 for _ in range(len(window_shape))), spec_val, out_channels),
				dtype=param.dtype,
			)
			b = (
				jax.random.normal(
					rng, (*window_shape, in_channels, spec_val), dtype=param.dtype
				)
				* stddev
			)
			return LoraWeight(param, a, b, alpha=alpha)

		return jax.tree_util.tree_map_with_path(
			get_param,
			param_tree,
			lora_spec,
			is_leaf=is_leaf,
		)

	@staticmethod
	def wrap_tx(
		tx: optax.GradientTransformation,
		lora_spec: Union[dict, jax.tree_util.PyTreeDef],
		scalar_frozen_grads: bool = False,
	) -> optax.GradientTransformation:
		"""Wraps an optimizer to handle LoRA parameters and freezing.

		This function customizes an Optax gradient transformation
		(`optax.GradientTransformation`) to work seamlessly with LoRA and
		parameter freezing.  It first creates a mask based on `lora_spec`,
		freezing parameters marked with `LORA_FREEZE`. Then it freezes the 'w'
		attribute of all `LoraWeight` instances, preventing the original
		weights from being updated.

		Args:
		    tx: The base Optax gradient transformation to wrap.
		    lora_spec: A PyTree specifying the fine-tuning strategy for each
		        parameter.
		    scalar_frozen_grads: If True, frozen gradients will be replaced
		        with scalar zeros instead of zero-filled arrays. This can
		        improve memory efficiency.

		Returns:
		    The wrapped Optax gradient transformation that handles LoRA
		    parameters and parameter freezing.
		"""
		full_freeze_labels = jax.tree_util.tree_map(
			lambda x: "freeze" if x == LORA_FREEZE else "train",
			lora_spec,
		)
		optimizer_with_full_freeze = freeze_subtrees(
			tx,
			full_freeze_labels,
			use_scalar_zeros=scalar_frozen_grads,
		)
		return freeze_keys(
			optimizer_with_full_freeze,
			LoraWeight,
			"w",
			use_scalar_zeros=scalar_frozen_grads,
		)

	def apply_lora(
		self,
		module: Union[Any, flax.linen.Module],
		parameters: Union[dict, flax.core.FrozenDict],
		tx: optax.GradientTransformation,
		decision_fn: Optional[Callable] = None,
		tune_vectors: bool = False,
		rng: Optional[jax.random.PRNGKey] = None,
		stddev: float = 0.01,
		alpha: float = 1.0,
		is_leaf: bool = None,
		func_target: Optional[str] = None,
	) -> RaptureModule:
		"""Applies LoRA to a Flax module and wraps the optimizer.

		This function applies LoRA to the provided Flax module based on the
		`RaptureConfig` and the given `decision_fn`. It initializes LoRA
		parameters, wraps the optimizer to handle LoRA and frozen parameters,
		and returns an `XRaptureModule` containing the adapted module and
		related data.

		Args:
		    module: The Flax module to apply LoRA to.
		    parameters: The PyTree of parameters from the Flax module.
		    tx: The Optax gradient transformation (optimizer) to use.
		    decision_fn: An optional callable to override the default
		        `base_decision_function` for determining parameter-specific
		        fine-tuning strategies.
		    tune_vectors: Whether to tune vector parameters using full fine-tuning.
		    rng: A JAX PRNG key for random initialization.
		    stddev: Standard deviation for initializing LoRA adapter matrices.
		    alpha: Scaling factor for the LoRA adapters.
		    is_leaf: Optional function to determine leaf nodes in PyTrees.

		Returns:
		    An `XRaptureModule` containing the LoRA-adapted module, optimizer
		    state, parameters, and specifications.
		"""
		if rng is None:
			rng = jax.random.PRNGKey(0)
		lora_spec = self.make_lora_specs(
			parameters=parameters,
			tune_vectors=tune_vectors,
			decision_fn=decision_fn,
		)

		lora_parameters = self.init_lora_parameters(
			param_tree=parameters,
			dtype=self.config.dtype,
			lora_spec=lora_spec,
			rng=rng,
			stddev=stddev,
			alpha=alpha,
			is_leaf=is_leaf,
		)
		tx = self.wrap_tx(tx=tx, lora_spec=lora_spec)
		opt_state = tx.init(lora_parameters)
		if func_target is None:
			lora_model = implicit_compact(module)
		else:
			lora_model = implicit_compact(getattr(module, func_target))
		return RaptureModule(
			lora_opt_state=opt_state,
			lora_module=lora_model,
			lora_specs=lora_spec,
			lora_parameters=lora_parameters,
			lora_tx=tx,
		)



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for common pruners."""

import functools

import chex
import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized

from fjformer.jaxpruner import sparsity_distributions, utils
from fjformer.jaxpruner.algorithms import pruners


class PrunerTest(parameterized.TestCase, absltest.TestCase):
	def testMagnitudePruning(self):
		updater = pruners.MagnitudePruning()
		param_tree = {"w": jnp.array([0.1, 0.9, -0.2, 0.5])}
		expected = {"w": jnp.array([0.1, 0.9, 0.2, 0.5])}
		chex.assert_trees_all_close(updater.calculate_scores(param_tree), expected)

	def testSaliencyPruning(self):
		updater = pruners.SaliencyPruning()
		param_tree = {"w": jnp.array([0.1, 0.9, 0.2, 0.5])}
		grad_tree = {"w": jnp.array([5, 1, 3, 0.5])}
		expected = {"w": jnp.array([0.5, 0.9, 0.6, 0.25])}
		chex.assert_trees_all_close(
			updater.calculate_scores(params=param_tree, grads=grad_tree), expected
		)

	def testRandomPruning(self):
		target_sparsity = 0.5
		sparsity_fn = functools.partial(
			sparsity_distributions.uniform, sparsity=target_sparsity
		)
		updater = pruners.RandomPruning(sparsity_distribution_fn=sparsity_fn)
		param_tree = {"w": jax.random.normal(jax.random.PRNGKey(0), (10, 10))}
		updater.rng_seed = jax.random.PRNGKey(0)
		score1 = updater.calculate_scores(param_tree)
		updater.rng_seed = jax.random.PRNGKey(5)
		score2 = updater.calculate_scores(param_tree)
		self.assertFalse(
			jnp.array_equal(
				score1["w"],
				score2["w"],
			)
		)

		mask = updater.create_masks(score1, {"w": target_sparsity})
		self.assertEqual(utils.summarize_sparsity(mask)["_total_sparsity"], 0.5)


if __name__ == "__main__":
	absltest.main()



"""
fjformer core for manipulating jax.Arrays adapted from davisyoshida/qax
"""

from fjformer.core import symbols as symbols
from fjformer.core import types as types
from fjformer.core import utilities as utilities
from fjformer.core.implicit_array import ArrayValue as ArrayValue
from fjformer.core.implicit_array import EmptyNode as EmptyNode
from fjformer.core.implicit_array import ImplicitArray as ImplicitArray
from fjformer.core.implicit_array import UninitializedAval as UninitializedAval
from fjformer.core.implicit_array import aux_field as aux_field
from fjformer.core.implicit_array import default_handler as default_handler
from fjformer.core.implicit_array import implicit_compact as implicit_compact
from fjformer.core.implicit_array import materialize_nested as materialize_nested
from fjformer.core.implicit_array import primitive_handler as primitive_handler
from fjformer.core.implicit_array import (
	tree_flatten_with_implicit as tree_flatten_with_implicit,
	tree_flatten_with_path_with_implicit as tree_flatten_with_path_with_implicit,
	tree_leaves_with_implicit as tree_leaves_with_implicit,
	tree_map_with_implicit as tree_map_with_implicit,
	tree_map_with_path_with_implicit as tree_map_with_path_with_implicit,
	tree_structure_with_implicit as tree_structure_with_implicit,
)



import warnings
from typing import Optional, Tuple

import chex
import optax


def _get_adamw_base(
	scheduler: optax.Schedule,
	b1: float = 0.9,
	b2: float = 0.999,
	eps: float = 1e-8,
	eps_root: float = 0.0,
	weight_decay: float = 1e-1,
	gradient_accumulation_steps: int = 1,
	mu_dtype: Optional[chex.ArrayDType] = None,  # type:ignore
	clip_grad: Optional[float] = None,
	**kwargs,
) -> optax.GradientTransformation:
	"""
	Creates a base AdamW optimizer with the given scheduler.

	Args:
	    scheduler (optax.Schedule): Learning rate scheduler.
	    b1 (float): Adam beta1 parameter.
	    b2 (float): Adam beta2 parameter.
	    eps (float): Adam epsilon parameter.
	    eps_root (float): Adam epsilon root parameter.
	    weight_decay (float): Weight decay factor.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    mu_dtype (Optional[chex.ArrayDType]): Data type for momentum.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    optax.GradientTransformation: The configured optimizer.
	"""
	for kwarg in kwargs.keys():
		warnings.warn(f"Key {kwarg} is not used for optimizer.", stacklevel=1)
	chain = [
		optax.scale_by_adam(
			b1=b1,
			b2=b2,
			eps=eps,
			eps_root=eps_root,
			mu_dtype=mu_dtype,
		),
		optax.add_decayed_weights(weight_decay=weight_decay),
		optax.scale_by_schedule(scheduler),
		optax.scale(-1),
	]

	if clip_grad is not None:
		chain.insert(0, optax.clip_by_global_norm(clip_grad))

	tx = optax.chain(*chain)

	if gradient_accumulation_steps > 1:
		tx = optax.MultiSteps(tx, gradient_accumulation_steps)

	return tx


def get_adamw_with_cosine_scheduler(
	steps: int,
	learning_rate: float = 5e-5,
	b1: float = 0.9,
	b2: float = 0.999,
	eps: float = 1e-8,
	eps_root: float = 0.0,
	weight_decay: float = 1e-1,
	gradient_accumulation_steps: int = 1,
	mu_dtype: Optional[chex.ArrayDType] = None,  # type:ignore
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an AdamW optimizer with a cosine learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate (float): Initial learning rate.
	    b1 (float): Adam beta1 parameter.
	    b2 (float): Adam beta2 parameter.
	    eps (float): Adam epsilon parameter.
	    eps_root (float): Adam epsilon root parameter.
	    weight_decay (float): Weight decay factor.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    mu_dtype (Optional[chex.ArrayDType]): Data type for momentum.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.cosine_decay_schedule(init_value=learning_rate, decay_steps=steps)
	tx = _get_adamw_base(
		scheduler=scheduler,
		b1=b1,
		b2=b2,
		eps=eps,
		eps_root=eps_root,
		weight_decay=weight_decay,
		gradient_accumulation_steps=gradient_accumulation_steps,
		mu_dtype=mu_dtype,
		clip_grad=clip_grad,
		**kwargs,
	)
	return tx, scheduler


def get_adamw_with_linear_scheduler(
	steps: int,
	learning_rate_start: float = 5e-5,
	learning_rate_end: float = 1e-5,
	b1: float = 0.9,
	b2: float = 0.999,
	eps: float = 1e-8,
	eps_root: float = 0.0,
	weight_decay: float = 1e-1,
	gradient_accumulation_steps: int = 1,
	mu_dtype: Optional[chex.ArrayDType] = None,  # type:ignore
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an AdamW optimizer with a linear learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate_start (float): Initial learning rate.
	    learning_rate_end (float): Final learning rate.
	    b1 (float): Adam beta1 parameter.
	    b2 (float): Adam beta2 parameter.
	    eps (float): Adam epsilon parameter.
	    eps_root (float): Adam epsilon root parameter.
	    weight_decay (float): Weight decay factor.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    mu_dtype (Optional[chex.ArrayDType]): Data type for momentum.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.linear_schedule(
		init_value=learning_rate_start,
		end_value=learning_rate_end,
		transition_steps=steps,
	)
	tx = _get_adamw_base(
		scheduler=scheduler,
		b1=b1,
		b2=b2,
		eps=eps,
		eps_root=eps_root,
		weight_decay=weight_decay,
		gradient_accumulation_steps=gradient_accumulation_steps,
		mu_dtype=mu_dtype,
		clip_grad=clip_grad,
		**kwargs,
	)
	return tx, scheduler


def get_adamw_with_warmup_linear_scheduler(
	steps: int,
	learning_rate_start: float = 5e-5,
	learning_rate_end: float = 1e-5,
	b1: float = 0.9,
	b2: float = 0.999,
	eps: float = 1e-8,
	eps_root: float = 0.0,
	weight_decay: float = 1e-1,
	gradient_accumulation_steps: int = 1,
	mu_dtype: Optional[chex.ArrayDType] = None,  # type:ignore
	warmup_steps: int = 100,
	warmup_init_value: float = 5e-8,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an AdamW optimizer with a linear learning rate scheduler and warm-up phase.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate_start (float): Learning rate after warm-up.
	    learning_rate_end (float): Final learning rate.
	    b1 (float): Adam beta1 parameter.
	    b2 (float): Adam beta2 parameter.
	    eps (float): Adam epsilon parameter.
	    eps_root (float): Adam epsilon root parameter.
	    weight_decay (float): Weight decay factor.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    mu_dtype (Optional[chex.ArrayDType]): Data type for momentum.
	    warmup_steps (int): Number of warm-up steps.
	    warmup_init_value (float): Initial learning rate for warm-up.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler_warmup = optax.linear_schedule(
		init_value=warmup_init_value,
		end_value=learning_rate_start,
		transition_steps=warmup_steps,
	)
	scheduler_decay = optax.linear_schedule(
		init_value=learning_rate_start,
		end_value=learning_rate_end,
		transition_steps=steps - warmup_steps,
	)

	scheduler_combined = optax.join_schedules(
		schedules=[scheduler_warmup, scheduler_decay], boundaries=[warmup_steps]
	)

	tx = _get_adamw_base(
		scheduler=scheduler_combined,
		b1=b1,
		b2=b2,
		eps=eps,
		eps_root=eps_root,
		weight_decay=weight_decay,
		gradient_accumulation_steps=gradient_accumulation_steps,
		mu_dtype=mu_dtype,
		clip_grad=clip_grad,
		**kwargs,
	)
	return tx, scheduler_combined


def get_adamw_with_warmup_cosine_scheduler(
	steps: int,
	learning_rate: float = 5e-5,
	learning_rate_end: float = 1e-5,
	b1: float = 0.9,
	b2: float = 0.999,
	eps: float = 1e-8,
	eps_root: float = 0.0,
	weight_decay: float = 1e-1,
	exponent: float = 1.0,
	gradient_accumulation_steps: int = 1,
	warmup_steps: int = 100,
	mu_dtype: Optional[chex.ArrayDType] = None,  # type:ignore
	warmup_init_value: float = 0.5e-7,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an AdamW optimizer with a cosine learning rate scheduler and warm-up phase.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate (float): Peak learning rate after warm-up.
	    learning_rate_end (float): Final learning rate.
	    b1 (float): Adam beta1 parameter.
	    b2 (float): Adam beta2 parameter.
	    eps (float): Adam epsilon parameter.
	    eps_root (float): Adam epsilon root parameter.
	    weight_decay (float): Weight decay factor.
	    exponent (float): Exponent for the cosine decay.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    warmup_steps (int): Number of warm-up steps.
	    mu_dtype (Optional[chex.ArrayDType]): Data type for momentum.
	    warmup_init_value (float): Initial learning rate for warm-up.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.warmup_cosine_decay_schedule(
		init_value=warmup_init_value,
		peak_value=learning_rate,
		warmup_steps=warmup_steps,
		decay_steps=steps,
		end_value=learning_rate_end,
		exponent=exponent,
	)
	tx = _get_adamw_base(
		scheduler=scheduler,
		b1=b1,
		b2=b2,
		eps=eps,
		eps_root=eps_root,
		weight_decay=weight_decay,
		gradient_accumulation_steps=gradient_accumulation_steps,
		mu_dtype=mu_dtype,
		clip_grad=clip_grad,
		**kwargs,
	)
	return tx, scheduler



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains apis to use the JaxPruner libraries."""

import copy
import functools
import logging

import optax

from fjformer.jaxpruner import (
	algorithms,
	base_updater,
	sparsity_distributions,
	sparsity_schedules,
	sparsity_types,
)

ALGORITHM_REGISTRY = {
	"no_prune": base_updater.NoPruning,
	"magnitude": algorithms.MagnitudePruning,
	"random": algorithms.RandomPruning,
	"saliency": algorithms.SaliencyPruning,
	"magnitude_ste": algorithms.SteMagnitudePruning,
	"random_ste": algorithms.SteRandomPruning,
	"global_magnitude": algorithms.GlobalMagnitudePruning,
	"global_saliency": algorithms.GlobalSaliencyPruning,
	"static_sparse": algorithms.StaticRandomSparse,
	"rigl": algorithms.RigL,
	"set": algorithms.SET,
}
ALGORITHMS = tuple(ALGORITHM_REGISTRY.keys())


def all_algorithm_names():
	"""Returns all algorithm names registered."""
	return tuple(ALGORITHM_REGISTRY.keys())


def register_algorithm(algorithm_name, algorithm):
	"""Registers a new algorithm."""
	ALGORITHM_REGISTRY[algorithm_name] = algorithm


def create_updater_from_config(
	sparsity_config,
):
	"""Gets a sparsity updater based on the given sparsity config.

	A sample usage of this api is in below.
	```
	  sparsity_config = ml_collections.ConfigDict()
	  # Required
	  sparsity_config.algorithm = 'magnitude'
	  sparsity_config.dist_type = 'erk'
	  sparsity_config.sparsity = 0.8

	  # Optional
	  sparsity_config.update_freq = 10
	  sparsity_config.update_end_step = 1000
	  sparsity_config.update_start_step = 200
	  sparsity_config.sparsity_type = 'nm_2,4'

	  updater = create_updater_from_config(sparsity_config)
	```

	- `algorithm`: str, one of jaxpruner.all_algorithm_names()
	- `dist_type`: str, 'erk' or 'uniform'.
	- `update_freq`: int, passed to PeriodicSchedule.
	- `update_end_step`: int, passed to PeriodicSchedule.
	- `update_start_step`: int, if None or doesn't exist NoUpdateSchedule is
	  used. If equal to `update_end_step`, OneShotSchedule is used. Otherwise
	  PolynomialSchedule is used.
	- `sparsity`: str, float or jaxpruner.SparsityType, if float, then
	  SparsityType.Unstructured is used. If str in '{N}:{M}' format,
	  then SparsityType.NbyM is used. If str in '{N}x{M}' format,
	  then SparsityType.Block is used. User can also pass the desired
	  SparsityType directly. Everything else under `sparsity_config`
	  is passed to the algorithm directly.

	Args:
	  sparsity_config: configuration for the algorithm. See options above.

	Returns:
	  an algorithm with the given configuration.
	"""
	logging.info("Creating  updater for %s", sparsity_config.algorithm)
	if sparsity_config.algorithm == "no_prune":
		return base_updater.NoPruning()

	config = copy.deepcopy(sparsity_config).unlock()

	if config.dist_type == "uniform":
		config.sparsity_distribution_fn = sparsity_distributions.uniform
	elif config.dist_type == "erk":
		config.sparsity_distribution_fn = sparsity_distributions.erk
	else:
		raise ValueError(
			f"dist_type: {config.dist_type} is not supported. " "Use `erk` or `uniform`"
		)
	del config.dist_type

	if config.get("filter_fn", None):
		if not config.algorithm.startswith("global_"):
			new_fn = functools.partial(
				config.sparsity_distribution_fn, filter_fn=config.filter_fn
			)
			config.sparsity_distribution_fn = new_fn
			del config.filter_fn

	if config.get("custom_sparsity_map", None):
		if not config.algorithm.startswith("global_"):
			new_fn = functools.partial(
				config.sparsity_distribution_fn,
				custom_sparsity_map=config.custom_sparsity_map,
			)
			config.sparsity_distribution_fn = new_fn
			del config.custom_sparsity_map

	if config.algorithm.startswith("global_"):
		# Distribution function is not used.
		del config.sparsity_distribution_fn
	else:
		kwargs = {"sparsity": config.sparsity}
		del config.sparsity
		if config.get("filter_fn", None):
			kwargs["filter_fn"] = config.filter_fn
			del config.filter_fn
		config.sparsity_distribution_fn = functools.partial(
			config.sparsity_distribution_fn, **kwargs
		)
	if config.get("sparsity_type", None):
		s_type = config.sparsity_type
		if isinstance(s_type, str) and s_type.startswith("nm"):
			# example: nm_2,4
			n, m = s_type.split("_")[1].strip().split(",")
			del config.sparsity_type
			config.sparsity_type = sparsity_types.NByM(int(n), int(m))
		elif isinstance(s_type, str) and (s_type.startswith("block")):
			# example: block_4,4
			n, m = s_type.split("_")[1].strip().split(",")
			del config.sparsity_type
			config.sparsity_type = sparsity_types.Block(block_shape=(int(n), int(m)))
		elif isinstance(s_type, str) and (s_type.startswith("channel")):
			axis = int(s_type.split("_")[1])
			del config.sparsity_type
			config.sparsity_type = sparsity_types.Channel(axis=axis)
		else:
			raise ValueError(f"Sparsity type {s_type} is not supported.")

	if config.algorithm in ALGORITHM_REGISTRY:
		updater_type = ALGORITHM_REGISTRY[config.algorithm]
		if config.algorithm in ("rigl", "set"):
			config.drop_fraction_fn = optax.cosine_decay_schedule(
				config.get("drop_fraction", 0.1), config.update_end_step
			)
		del config.algorithm
	else:
		raise ValueError(
			f"Sparsity algorithm {config.algorithm} is not supported."
			" Please use a key from fjformer.jaxpruner.all_algorithm_names() or register"
			" the new algorithm using jaxpruner.register_algorithm()."
		)

	if config.get("update_start_step", None) is None:
		config.scheduler = sparsity_schedules.NoUpdateSchedule()
	elif config.update_end_step == config.update_start_step:
		config.scheduler = sparsity_schedules.OneShotSchedule(
			target_step=config.update_end_step
		)
	else:
		config.scheduler = sparsity_schedules.PolynomialSchedule(
			update_freq=config.update_freq,
			update_start_step=config.update_start_step,
			update_end_step=config.update_end_step,
		)
	for field_name in (
		"update_freq",
		"update_start_step",
		"update_end_step",
		"drop_fraction",
	):
		if hasattr(config, field_name):
			delattr(config, field_name)

	updater = updater_type(**config)

	return updater



from dataclasses import dataclass
from functools import partial
from typing import Any, Optional, Tuple

import jax
import numpy as np
from jax import lax
from jax import numpy as jnp

from fjformer.core.errors import (
	MaterializationError,
	OperationError,
	ShapeDtypeError,
	UnsupportedPrimitiveError,
)
from fjformer.core.implicit_array import (
	ELEMENTWISE_BINOPS,
	ELEMENTWISE_UNOPS,
	ArrayValue,
	ImplicitArray,
	aux_field,
	default_handler,
	implicit_compact,
	primitive_handler,
)
from fjformer.core.types import Complement

INFINITY = float("inf")
NEGATIVE_INFINITY = float("-inf")
_GENERAL = -2
_SPECIALIZED = -1


def _get_shape_dtype(
	x: Any, shape: Optional[Tuple[int, ...]] = None, dtype: Optional[Any] = None
) -> Tuple[Tuple[int, ...], Any]:
	"""
	Determine the shape and dtype for a given input.

	Args:
	    x: Input array or value
	    shape: Optional shape to use
	    dtype: Optional dtype to use

	Returns:
	    A tuple containing:
	        - shape: Tuple[int, ...]: The determined shape
	        - dtype: Any: The determined dtype

	Raises:
	    ShapeDtypeError: If the input x is None and both shape and dtype are None
	"""
	if x is None and shape is None and dtype is None:
		raise ShapeDtypeError("Either x or both shape and dtype must be provided")

	try:
		if shape is None:
			shape = np.shape(x)
		else:
			shape = jax.core.canonicalize_shape(shape)

		if dtype is None:
			dtype = lax.dtype(x)

		return shape, dtype
	except Exception as e:
		raise ShapeDtypeError(f"Error determining shape and dtype: {str(e)}") from e


def _out_shape_dtype(
	primitive: Any,
	*args: Any,
	**kwargs: Any,
):
	"""
	Determine the output shape and dtype for a given primitive operation.

	This function uses JAX's shape inference capabilities to determine what
	the shape and dtype of the output would be if the operation were performed
	on concrete arrays.

	Args:
	    primitive (Any): JAX primitive to be evaluated.
	    *args: Positional arguments for the primitive.
	    **kwargs: Keyword arguments for the primitive.

	Returns:
	    Tuple[Tuple[int, ...], Any]: A tuple containing:
	        - The shape of the output as a tuple of integers.
	        - The dtype of the output.
	"""
	out_aval = jax.eval_shape(
		partial(default_handler, primitive, **kwargs),
		*(jax.core.get_aval(x) for x in args),
	)
	return jax.tree_util.tree_map(lambda x: (x.shape, x.dtype), out_aval)


def symbolic_zero_like(
	x: Any, shape: Optional[Tuple[int, ...]] = None, dtype: Optional[Any] = None
) -> "SymbolicConstant":
	"""
	Create a SymbolicConstant filled with zeros, similar to the input.

	Args:
	    x: Input to base the result on
	    shape: Optional shape for the result
	    dtype: Optional dtype for the result

	Returns:
	    SymbolicConstant: A constant array filled with zeros

	Raises:
	    OperationError: If creation of the zero-filled SymbolicConstant fails
	"""
	try:
		dtype = lax.dtype(x) if dtype is None else dtype
		return symbolic_full_like(x, 0, shape=shape, dtype=dtype)
	except Exception as e:
		raise OperationError(
			f"Failed to create zero-filled SymbolicConstant: {str(e)}"
		) from e


def symbolic_full_like(
	x: Any,
	fill_value: Any,
	shape: Optional[Tuple[int, ...]] = None,
	dtype: Optional[Any] = None,
) -> "SymbolicConstant":
	"""
	Create a SymbolicConstant filled with a specific value, similar to the input.

	Args:
	    x: Input to base the result on
	    fill_value: Value to fill the SymbolicConstant with
	    shape: Optional shape for the result
	    dtype: Optional dtype for the result

	Returns:
	    SymbolicConstant filled with the specified value
	"""
	shape, _ = _get_shape_dtype(x, shape, None)
	if dtype is None:
		dtype = jax.lax.dtype(fill_value)

	return SymbolicConstant(fill_value, shape=shape, dtype=dtype)


@dataclass
class SymbolicConstant(ImplicitArray):
	"""
	Represents a constant array symbolically, allowing for efficient operations at compile time.

	Attributes:
	    value (Any): The constant value of the array
	    weak_type (bool): Whether the constant has a weak type
	"""

	value: Any = aux_field()
	weak_type: bool = aux_field(default=False)

	def __post_init__(self):
		"""Initialize the SymbolicConstant after creation."""
		super().__post_init__()
		try:
			with jax.ensure_compile_time_eval():
				self.value = jnp.asarray(self.value, dtype=self.dtype)
		except Exception as e:
			raise OperationError(f"Failed to initialize SymbolicConstant: {str(e)}") from e

	def compute_dtype(self) -> Any:
		"""Compute the dtype of the constant."""
		return lax.dtype(self.value)

	def materialize(self) -> jnp.ndarray:
		"""
		Materialize the symbolic constant into a concrete array.

		Returns:
		    jnp.ndarray: A concrete JAX array

		Raises:
		    MaterializationError: If materialization fails
		"""
		try:
			return jnp.full(self.shape, self.value, dtype=self.dtype)
		except Exception as e:
			raise MaterializationError(
				f"Failed to materialize SymbolicConstant: {str(e)}"
			) from e

	def copy(self) -> "SymbolicConstant":
		"""
		Create a copy of the SymbolicConstant.

		Returns:
		    SymbolicConstant: A copy of the current instance

		Raises:
		    OperationError: If copying fails
		"""
		try:
			return jax.tree_util.tree_map(lambda x: x, self)
		except Exception as e:
			raise OperationError(f"Failed to copy SymbolicConstant: {str(e)}") from e


@implicit_compact
def broadcast_to(val, shape):
	return jnp.broadcast_to(val, shape)


@implicit_compact
def astype(val, dtype):
	return val.astype(dtype)


@primitive_handler(
	[
		"reshape",
		"broadcast_in_dim",
		"reduce_min",
		"reduce_max",
		"reduce_or",
		"reduce_and",
	]
)
def unchanged_value_op(
	primitive: Any,
	sym: SymbolicConstant,
	**kwargs: Any,
) -> SymbolicConstant:
	"""
	Handler for JAX primitives that don't change the value of a SymbolicConstant.

	This function handles operations that may change the shape or other properties
	of a SymbolicConstant, but not its fundamental value.

	Args:
	    primitive (Any): The JAX primitive being handled.
	    sym (core.symbols.SymbolicConstant): The symbolic constant being operated on.
	    **kwargs: Additional keyword arguments for the primitive operation.

	Returns:
	    core.symbols.SymbolicConstant: A new SymbolicConstant with potentially
	    updated shape and dtype, but the same value as the input.
	"""
	out_shape, out_dtype = _out_shape_dtype(primitive, sym, **kwargs)
	return SymbolicConstant(sym.value, shape=out_shape, dtype=out_dtype)


def _op_and_reshape(primitive, lhs, rhs, flip=False):
	"""
	Close over one arg so we can do math at tracing time, but let the other one get traced.
	"""
	if flip:
		lhs, rhs = (rhs, lhs)

	@implicit_compact
	def inner(arg):
		other = lhs
		if flip:
			arg, other = (other, arg)

		result = default_handler(primitive, arg, other)
		return result

	return inner(rhs)


def special_case_binop(
	name,
	identity=None,
	annihilator=None,
	flip=False,
):
	"""
	Create a special case handler for binary operations.
	"""
	lhs_type = SymbolicConstant
	rhs_type = Complement[ArrayValue, SymbolicConstant]
	if flip:
		lhs_type, rhs_type = rhs_type, lhs_type

	@primitive_handler(name, precedence=_SPECIALIZED)
	def handler(primitive, lhs: lhs_type, rhs: rhs_type, **kwargs):  # type:ignore
		out_shape, out_dtype = _out_shape_dtype(
			primitive,
			lhs,
			rhs,
			**kwargs,
		)
		with jax.ensure_compile_time_eval():
			if lhs.value == identity:
				return broadcast_to(astype(rhs, out_dtype), out_shape)

			if lhs.value == annihilator:
				return SymbolicConstant(lhs.value, shape=out_shape, dtype=out_dtype)

			print(f"{primitive} {lhs.value} {rhs}")
			return _op_and_reshape(primitive, lhs.value, rhs)


# Define special case handlers for various binary operations
special_case_binop("add", identity=0)
special_case_binop("mul", identity=1, annihilator=0)
special_case_binop("and", annihilator=0)
special_case_binop("or", identity=0)
special_case_binop("xor", identity=0)
special_case_binop("sub", identity=0, flip=True)
special_case_binop("div", identity=1, flip=True)
special_case_binop("exp", identity=1, flip=True)
special_case_binop("min", identity=float("inf"), annihilator=float("-inf"))
special_case_binop("max", identity=float("-inf"), annihilator=float("inf"))


def eval_default_handler(primitive, *args, **kwargs):
	"""
	Evaluate the default handler for a primitive at compile time.
	"""
	with jax.ensure_compile_time_eval():
		result = primitive.bind(*args, **kwargs)
	return result


@primitive_handler(ELEMENTWISE_UNOPS, precedence=_GENERAL)
def handle_unop(primitive, sym: SymbolicConstant, **kwargs):
	"""
	Handle unary operations on SymbolicConstants.

	Args:
	    primitive: The JAX primitive to handle
	    sym: The SymbolicConstant to operate on
	    **kwargs: Additional keyword arguments

	Returns:
	    SymbolicConstant: The result of the unary operation

	Raises:
	    UnsupportedPrimitiveError: If the primitive is not supported
	    OperationError: If the operation fails
	"""

	try:
		new_val = eval_default_handler(primitive, sym.value, **kwargs)
		return symbolic_full_like(sym, new_val)
	except Exception as e:
		if isinstance(e, NotImplementedError):
			raise UnsupportedPrimitiveError(f"Unsupported primitive: {primitive}") from e
		raise OperationError(
			f"Failed to handle unary operation {primitive}: {str(e)}"
		) from e


@primitive_handler(ELEMENTWISE_BINOPS, precedence=_GENERAL)
def handle_binop(
	primitive,
	lhs: SymbolicConstant,
	rhs: SymbolicConstant,
	**kwargs,
):
	"""
	Handle binary operations on SymbolicConstants.
	"""
	out_shape, out_dtype = _out_shape_dtype(primitive, lhs, rhs, **kwargs)
	new_val = eval_default_handler(primitive, lhs.value, rhs.value, **kwargs)
	return symbolic_full_like(lhs, new_val, shape=out_shape, dtype=out_dtype)


@primitive_handler(["reduce_sum", "reduce_prod"])
def reduce_sum(
	primitive,
	sym: SymbolicConstant,
	*,
	axes,
):
	"""
	Handle reduction operations (sum and product) on SymbolicConstants.
	"""
	out_shape, out_dtype = _out_shape_dtype(primitive, sym, axes=axes)
	with jax.ensure_compile_time_eval():
		if sym.value == 0:
			return SymbolicConstant(0, shape=out_shape, dtype=out_dtype)

		orig_size = np.prod(sym.shape)
		new_size = np.prod(out_shape)

		n_combined = orig_size // new_size

		new_val = sym.value
		if primitive.name == "reduce_sum":
			new_val = new_val * n_combined
		else:
			new_val = new_val**n_combined

	return SymbolicConstant(new_val, shape=out_shape, dtype=out_dtype)


@primitive_handler("select_n")
def handle_select_n(
	primitive,
	cond_val,
	*arg_vals: SymbolicConstant,
):
	"""
	Handle select_n operation on SymbolicConstants.
	"""
	if len(set(val.value.item() for val in arg_vals)) != 1:
		return NotImplemented

	out_shape, out_dtype = _out_shape_dtype(primitive, cond_val, *arg_vals)
	return SymbolicConstant(arg_vals[0].value, shape=out_shape, dtype=out_dtype)



# Copyright 2023 The EASYDEL Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation by @erfanzar,
# with a few bug fixes and adjustments.

from functools import partial
from typing import Optional

import jax
import jax.interpreters
import jax.interpreters.pxla
import jax.random
from fjformer import GenerateRNG
from jax import numpy as jnp
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

PLATFORM = jax.extend.backend.get_backend().platform
INTERPRET = PLATFORM == "cpu"
rng = GenerateRNG()


def get_best_block_size_tpu(A, B):
	# A is assumed to be of shape (m, k) and B is of shape (k, n)
	m, k = A.shape[0], A.shape[1]
	n = B.shape[1]

	# Initialize block sizes
	bm, bk, bn = 16, 16, 16

	# Adjust block size for m
	if m >= 1024:
		bm = 1024

	# Adjust block size for k
	if k > 128:
		bk = 128
	if k >= 2048:
		bk = 1024

	# Adjust block size for n
	if n >= 1024:
		bn = 1024
	if n >= 2048:
		bn = 256

	return bm, bk, bn


def get_best_block_size(A, B):
	# A is assumed to be of shape (m, k) and B is of shape (k, n)
	m, k = A.shape[0], A.shape[1]
	n = B.shape[1]

	# Initialize block sizes
	bm, bk, bn = 16, 16, 16

	# Adjust block size for m
	if m > 1000:
		bm = 32
	if m > 2000:
		bm = 64
	if m > 8000:
		bm = 128

	# Adjust block size for k
	if k > 1000:
		bk = 32
	if k > 2000:
		bk = 64
	if k > 8000:
		bk = 128

	# Adjust block size for n
	if n > 1000:
		bn = 32  # bn maxes out at 32

	return bm, bk, bn


def _tpu_matmul_kernel_fwd(a_ref, b_ref, o_ref, ac_ref, *, precision, k_grid):
	@pl.when(pl.program_id(2) == 0)
	def _():
		ac_ref[...] = jnp.zeros_like(ac_ref)

	ac_ref[...] += jnp.dot(
		a_ref[...].astype(jnp.float32),
		b_ref[...].astype(jnp.float32),
		preferred_element_type=jnp.float32,
		precision=precision,
	)

	@pl.when(pl.program_id(2) == k_grid - 1)
	def _():
		o_ref[...] = ac_ref[...].astype(o_ref.dtype)


@partial(
	jax.jit,
	static_argnames=[
		"blocksize_m",
		"blocksize_k",
		"blocksize_n",
		"precision",
	],
)
def _call_tpu_matmul_kernel_fwd(
	A: jax.Array,
	B: jax.Array,
	*,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	precision: jax.lax.PrecisionLike = None,
):
	assert A.ndim == 2 and B.ndim == 2, f"got {A.shape=} and {B.shape=}"
	assert (
		A.shape[1] == B.shape[0]
	), f"matmul can't be operated with these shapes {A.shape=} {B.shape=} "
	m, k, n = A.shape[0], B.shape[0], B.shape[1]
	pbm, pbk, pbn = get_best_block_size(A, B)

	blocksize_m = blocksize_m or pbm
	blocksize_n = blocksize_n or pbn
	blocksize_k = blocksize_k or pbk

	grid = (
		pl.cdiv(m, blocksize_m),
		pl.cdiv(n, blocksize_n),
		pl.cdiv(k, blocksize_k),
	)

	in_specs = [
		pl.BlockSpec((blocksize_m, blocksize_k), lambda mi, ni, ki: (mi, ki)),
		pl.BlockSpec((blocksize_k, blocksize_n), lambda mi, ni, ki: (ki, ni)),
	]

	out_specs = pl.BlockSpec((blocksize_m, blocksize_n), lambda mi, ni, ki: (mi, ni))
	return pl.pallas_call(
		f=partial(_tpu_matmul_kernel_fwd, precision=precision, k_grid=grid[-1]),
		out_shape=jax.ShapeDtypeStruct(shape=(m, n), dtype=A.dtype),
		debug=False,
		interpret=INTERPRET,
		grid_spec=pltpu.PrefetchScalarGridSpec(
			num_scalar_prefetch=0,
			grid=grid,
			in_specs=in_specs,
			out_specs=out_specs,
			scratch_shapes=[pltpu.VMEM((blocksize_m, blocksize_n), jnp.float32)],
		),
		compiler_params=dict(
			mosaic=dict(dimension_semantics=("parallel", "parallel", "arbitrary"))
		),
		name="tpu_matmul_kernel_fwd",
	)(A, B)


def _call_tpu_matmul_kernel_fwd_residual(
	A: jax.Array,
	B: jax.Array,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	precision: jax.lax.PrecisionLike = None,
):
	return _call_tpu_matmul_kernel_fwd(
		A=A,
		B=B,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		precision=precision,
	), (A, B)


def _call_tpu_matmul_kernel_bwd(
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	precision: jax.lax.PrecisionLike,
	res,
	gO,
):
	(A, B) = res
	gA = _call_tpu_matmul_kernel_fwd(
		A=gO,
		B=B.transpose(1, 0),
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		precision=precision,
	)

	gB = _call_tpu_matmul_kernel_fwd(
		A=A.transpose(1, 0),
		B=gO,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		precision=precision,
	)
	return gA, gB


@partial(jax.custom_vjp, nondiff_argnums=(2, 3, 4, 5))
def pallas_gemm(
	A: jax.Array,
	B: jax.Array,
	blocksize_m: Optional[int],
	blocksize_k: Optional[int],
	blocksize_n: Optional[int],
	precision: jax.lax.PrecisionLike = None,
):
	return _call_tpu_matmul_kernel_fwd(
		A=A,
		B=B,
		blocksize_m=blocksize_m,
		blocksize_k=blocksize_k,
		blocksize_n=blocksize_n,
		precision=precision,
	)


pallas_gemm.defvjp(_call_tpu_matmul_kernel_fwd_residual, _call_tpu_matmul_kernel_bwd)

__all__ = ["pallas_gemm"]



from typing import Any, Optional, Tuple

import chex
import jax.numpy as jnp
import optax

from fjformer.optimizers.optimizer_utils import optax_add_scheduled_weight_decay
import warnings


def _get_adafactor_base(
	scheduler: optax.Schedule,
	min_dim_size_to_factor: int = 128,
	decay_rate: float = 0.8,
	decay_offset: int = 0,
	multiply_by_parameter_scale: bool = True,
	clipping_threshold: Optional[float] = 1.0,
	momentum: Optional[float] = None,
	dtype_momentum: chex.ArrayDType = jnp.float32,  # type:ignore
	weight_decay_rate: Optional[float] = None,
	eps: float = 1e-30,
	factored: bool = True,
	weight_decay: float = 0.0,
	weight_decay_mask: Optional[Any] = None,
	gradient_accumulation_steps: int = 1,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> optax.GradientTransformation:
	"""
	Creates a base Adafactor optimizer with the given scheduler and options.

	Args:
	    scheduler (optax.Schedule): Learning rate scheduler.
	    min_dim_size_to_factor (int): Minimum dimension size for factoring.
	    decay_rate (float): Decay rate for moment estimates.
	    decay_offset (int): Offset for decay calculations.
	    multiply_by_parameter_scale (bool): Whether to scale updates by parameter scale.
	    clipping_threshold (Optional[float]): Gradient clipping threshold.
	    momentum (Optional[float]): Momentum factor.
	    dtype_momentum (chex.ArrayDType): Data type for momentum.
	    weight_decay_rate (Optional[float]): Weight decay rate for Adafactor.
	    eps (float): Epsilon for numerical stability.
	    factored (bool): Whether to use factored second moment estimates.
	    weight_decay (float): Additional weight decay factor.
	    weight_decay_mask (Optional[Any]): Mask for weight decay.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    optax.GradientTransformation: The configured optimizer.
	"""

	for kwarg in kwargs.keys():
		warnings.warn(f"Key {kwarg} is not used for optimizer.", stacklevel=1)
	chain = [
		optax.adafactor(
			learning_rate=scheduler,
			min_dim_size_to_factor=min_dim_size_to_factor,
			decay_rate=decay_rate,
			decay_offset=decay_offset,
			multiply_by_parameter_scale=multiply_by_parameter_scale,
			clipping_threshold=clipping_threshold,
			eps=eps,
			momentum=momentum,
			weight_decay_rate=weight_decay_rate,
			dtype_momentum=dtype_momentum,
			factored=factored,
		)
	]
	if clip_grad is not None:
		chain.insert(0, optax.clip_by_global_norm(clip_grad))

	if weight_decay != 0.0:
		chain.append(
			optax_add_scheduled_weight_decay(
				lambda step: -scheduler(step) * weight_decay, weight_decay_mask
			)
		)

	tx = optax.chain(*chain)

	if gradient_accumulation_steps > 1:
		tx = optax.MultiSteps(tx, gradient_accumulation_steps)

	return tx


def get_adafactor_with_linear_scheduler(
	steps: int,
	learning_rate_start: float = 5e-5,
	learning_rate_end: float = 1e-5,
	weight_decay: float = 1e-1,
	min_dim_size_to_factor: int = 128,
	decay_rate: float = 0.8,
	decay_offset: int = 0,
	multiply_by_parameter_scale: bool = True,
	clipping_threshold: Optional[float] = 1.0,
	momentum: Optional[float] = None,
	dtype_momentum: chex.ArrayDType = jnp.float32,  # type:ignore
	weight_decay_rate: Optional[float] = None,
	eps: float = 1e-30,
	factored: bool = True,
	gradient_accumulation_steps: int = 1,
	weight_decay_mask: Optional[Any] = None,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an Adafactor optimizer with a linear learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate_start (float): Initial learning rate.
	    learning_rate_end (float): Final learning rate.
	    weight_decay (float): Weight decay factor.
	    min_dim_size_to_factor (int): Minimum dimension size for factoring.
	    decay_rate (float): Decay rate for moment estimates.
	    decay_offset (int): Offset for decay calculations.
	    multiply_by_parameter_scale (bool): Whether to scale updates by parameter scale.
	    clipping_threshold (Optional[float]): Gradient clipping threshold.
	    momentum (Optional[float]): Momentum factor.
	    dtype_momentum (chex.ArrayDType): Data type for momentum.
	    weight_decay_rate (Optional[float]): Weight decay rate for Adafactor.
	    eps (float): Epsilon for numerical stability.
	    factored (bool): Whether to use factored second moment estimates.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    weight_decay_mask (Optional[Any]): Mask for weight decay.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.linear_schedule(
		init_value=learning_rate_start,
		end_value=learning_rate_end,
		transition_steps=steps,
	)

	tx = _get_adafactor_base(
		scheduler=scheduler,
		min_dim_size_to_factor=min_dim_size_to_factor,
		decay_rate=decay_rate,
		decay_offset=decay_offset,
		multiply_by_parameter_scale=multiply_by_parameter_scale,
		clipping_threshold=clipping_threshold,
		momentum=momentum,
		dtype_momentum=dtype_momentum,
		weight_decay_rate=weight_decay_rate,
		eps=eps,
		factored=factored,
		weight_decay=weight_decay,
		weight_decay_mask=weight_decay_mask,
		gradient_accumulation_steps=gradient_accumulation_steps,
		clip_grad=clip_grad,
		**kwargs,
	)

	return tx, scheduler


def get_adafactor_with_warmup_linear_scheduler(
	steps: int,
	learning_rate_start: float = 5e-5,
	learning_rate_end: float = 1e-5,
	warmup_steps: int = 100,
	min_dim_size_to_factor: int = 128,
	decay_rate: float = 0.8,
	decay_offset: int = 0,
	multiply_by_parameter_scale: bool = True,
	clipping_threshold: Optional[float] = 1.0,
	momentum: Optional[float] = None,
	dtype_momentum: chex.ArrayDType = jnp.float32,  # type:ignore
	weight_decay_rate: Optional[float] = None,
	eps: float = 1e-30,
	factored: bool = True,
	gradient_accumulation_steps: int = 1,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an Adafactor optimizer with a warm-up linear learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate_start (float): Learning rate after warm-up.
	    learning_rate_end (float): Final learning rate.
	    warmup_steps (int): Number of warm-up steps.
	    min_dim_size_to_factor (int): Minimum dimension size for factoring.
	    decay_rate (float): Decay rate for moment estimates.
	    decay_offset (int): Offset for decay calculations.
	    multiply_by_parameter_scale (bool): Whether to scale updates by parameter scale.
	    clipping_threshold (Optional[float]): Gradient clipping threshold.
	    momentum (Optional[float]): Momentum factor.
	    dtype_momentum (chex.ArrayDType): Data type for momentum.
	    weight_decay_rate (Optional[float]): Weight decay rate for Adafactor.
	    eps (float): Epsilon for numerical stability.
	    factored (bool): Whether to use factored second moment estimates.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler_warmup = optax.linear_schedule(
		init_value=5e-8, end_value=learning_rate_start, transition_steps=warmup_steps
	)
	scheduler_decay = optax.linear_schedule(
		init_value=learning_rate_start,
		end_value=learning_rate_end,
		transition_steps=steps - warmup_steps,
	)

	scheduler_combined = optax.join_schedules(
		schedules=[scheduler_warmup, scheduler_decay], boundaries=[warmup_steps]
	)

	tx = _get_adafactor_base(
		scheduler=scheduler_combined,
		min_dim_size_to_factor=min_dim_size_to_factor,
		decay_rate=decay_rate,
		decay_offset=decay_offset,
		multiply_by_parameter_scale=multiply_by_parameter_scale,
		clipping_threshold=clipping_threshold,
		momentum=momentum,
		dtype_momentum=dtype_momentum,
		weight_decay_rate=weight_decay_rate,
		eps=eps,
		factored=factored,
		gradient_accumulation_steps=gradient_accumulation_steps,
		clip_grad=clip_grad,
		**kwargs,
	)

	return tx, scheduler_combined


def get_adafactor_with_cosine_scheduler(
	steps: int,
	learning_rate: float = 5e-5,
	min_dim_size_to_factor: int = 128,
	decay_rate: float = 0.8,
	decay_offset: int = 0,
	multiply_by_parameter_scale: bool = True,
	clipping_threshold: Optional[float] = 1.0,
	momentum: Optional[float] = None,
	dtype_momentum: chex.ArrayDType = jnp.float32,  # type:ignore
	weight_decay_rate: Optional[float] = None,
	eps: float = 1e-30,
	factored: bool = True,
	gradient_accumulation_steps: int = 1,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an Adafactor optimizer with a cosine learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate (float): Peak learning rate.
	    min_dim_size_to_factor (int): Minimum dimension size for factoring.
	    decay_rate (float): Decay rate for moment estimates.
	    decay_offset (int): Offset for decay calculations.
	    multiply_by_parameter_scale (bool): Whether to scale updates by parameter scale.
	    clipping_threshold (Optional[float]): Gradient clipping threshold.
	    momentum (Optional[float]): Momentum factor.
	    dtype_momentum (chex.ArrayDType): Data type for momentum.
	    weight_decay_rate (Optional[float]): Weight decay rate for Adafactor.
	    eps (float): Epsilon for numerical stability.
	    factored (bool): Whether to use factored second moment estimates.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.cosine_decay_schedule(init_value=learning_rate, decay_steps=steps)

	tx = _get_adafactor_base(
		scheduler=scheduler,
		min_dim_size_to_factor=min_dim_size_to_factor,
		decay_rate=decay_rate,
		decay_offset=decay_offset,
		multiply_by_parameter_scale=multiply_by_parameter_scale,
		clipping_threshold=clipping_threshold,
		momentum=momentum,
		dtype_momentum=dtype_momentum,
		weight_decay_rate=weight_decay_rate,
		eps=eps,
		factored=factored,
		gradient_accumulation_steps=gradient_accumulation_steps,
		clip_grad=clip_grad,
		**kwargs,
	)

	return tx, scheduler


def get_adafactor_with_warmup_cosine_scheduler(
	steps: int,
	learning_rate: float = 5e-5,
	learning_rate_end: float = 1e-5,
	weight_decay: float = 1e-1,
	min_dim_size_to_factor: int = 128,
	decay_rate: float = 0.8,
	decay_offset: int = 0,
	multiply_by_parameter_scale: bool = True,
	clipping_threshold: Optional[float] = 1.0,
	momentum: Optional[float] = None,
	dtype_momentum: chex.ArrayDType = jnp.float32,  # type:ignore
	weight_decay_rate: Optional[float] = None,
	eps: float = 1e-30,
	factored: bool = True,
	exponent: float = 1.0,
	weight_decay_mask: Optional[Any] = None,
	gradient_accumulation_steps: int = 1,
	warmup_steps: int = 100,
	clip_grad: Optional[float] = None,
	**kwargs,
) -> Tuple[optax.GradientTransformation, optax.Schedule]:
	"""
	Creates an Adafactor optimizer with a warm-up cosine learning rate scheduler.

	Args:
	    steps (int): Total number of training steps.
	    learning_rate (float): Peak learning rate after warm-up.
	    learning_rate_end (float): Final learning rate.
	    weight_decay (float): Weight decay factor.
	    min_dim_size_to_factor (int): Minimum dimension size for factoring.
	    decay_rate (float): Decay rate for moment estimates.
	    decay_offset (int): Offset for decay calculations.
	    multiply_by_parameter_scale (bool): Whether to scale updates by parameter scale.
	    clipping_threshold (Optional[float]): Gradient clipping threshold.
	    momentum (Optional[float]): Momentum factor.
	    dtype_momentum (chex.ArrayDType): Data type for momentum.
	    weight_decay_rate (Optional[float]): Weight decay rate for Adafactor.
	    eps (float): Epsilon for numerical stability.
	    factored (bool): Whether to use factored second moment estimates.
	    exponent (float): Exponent for the cosine decay.
	    weight_decay_mask (Optional[Any]): Mask for weight decay.
	    gradient_accumulation_steps (int): Number of steps to accumulate gradients.
	    warmup_steps (int): Number of warm-up steps.
	    clip_grad (Optional[float]): If provided, gradients will be clipped to this maximum norm.

	Returns:
	    Tuple[optax.GradientTransformation, optax.Schedule]: The optimizer and scheduler.
	"""
	scheduler = optax.warmup_cosine_decay_schedule(
		init_value=0.5e-7,
		peak_value=learning_rate,
		warmup_steps=warmup_steps,
		decay_steps=steps,
		end_value=learning_rate_end,
		exponent=exponent,
	)

	tx = _get_adafactor_base(
		scheduler=scheduler,
		min_dim_size_to_factor=min_dim_size_to_factor,
		decay_rate=decay_rate,
		decay_offset=decay_offset,
		multiply_by_parameter_scale=multiply_by_parameter_scale,
		clipping_threshold=clipping_threshold,
		momentum=momentum,
		dtype_momentum=dtype_momentum,
		weight_decay_rate=weight_decay_rate,
		eps=eps,
		factored=factored,
		weight_decay=weight_decay,
		weight_decay_mask=weight_decay_mask,
		gradient_accumulation_steps=gradient_accumulation_steps,
		clip_grad=clip_grad,
		**kwargs,
	)

	return tx, scheduler



from fjformer.sharding.sharding import auto_namedsharding as auto_namedsharding
from fjformer.sharding.sharding import auto_partition_spec as auto_partition_spec
from fjformer.sharding.sharding import auto_shard_array as auto_shard_array
from fjformer.sharding.sharding import create_mesh as create_mesh
from fjformer.sharding.sharding import flatten_tree as flatten_tree
from fjformer.sharding.sharding import get_jax_mesh as get_jax_mesh
from fjformer.sharding.sharding import get_metrics as get_metrics
from fjformer.sharding.sharding import (
	get_names_from_partition_spec as get_names_from_partition_spec,
)
from fjformer.sharding.sharding import (
	make_shard_and_gather_fns as make_shard_and_gather_fns,
)
from fjformer.sharding.sharding import match_partition_rules as match_partition_rules
from fjformer.sharding.sharding import named_tree_map as named_tree_map
from fjformer.sharding.sharding import names_in_current_mesh as names_in_current_mesh
from fjformer.sharding.sharding import tree_apply as tree_apply
from fjformer.sharding.sharding import tree_path_to_string as tree_path_to_string
from fjformer.sharding.sharding import (
	with_sharding_constraint as with_sharding_constraint,
)



# Addapted from https://github.com/google-research/jaxpruner
# to make it available on pypi and use on Easydel.
# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""APIs for jaxpruner."""

from fjformer.jaxpruner.algorithms import SET as SET
from fjformer.jaxpruner.algorithms import (
	GlobalMagnitudePruning as GlobalMagnitudePruning,
)
from fjformer.jaxpruner.algorithms import GlobalSaliencyPruning as GlobalSaliencyPruning
from fjformer.jaxpruner.algorithms import MagnitudePruning as MagnitudePruning
from fjformer.jaxpruner.algorithms import RandomPruning as RandomPruning
from fjformer.jaxpruner.algorithms import RigL as RigL
from fjformer.jaxpruner.algorithms import SaliencyPruning as SaliencyPruning
from fjformer.jaxpruner.algorithms import StaticRandomSparse as StaticRandomSparse
from fjformer.jaxpruner.algorithms import SteMagnitudePruning as SteMagnitudePruning
from fjformer.jaxpruner.algorithms import SteRandomPruning as SteRandomPruning
from fjformer.jaxpruner.api import ALGORITHM_REGISTRY as ALGORITHM_REGISTRY
from fjformer.jaxpruner.api import all_algorithm_names as all_algorithm_names
from fjformer.jaxpruner.api import (
	create_updater_from_config as create_updater_from_config,
)
from fjformer.jaxpruner.api import register_algorithm as register_algorithm
from fjformer.jaxpruner.base_updater import BaseUpdater as BaseUpdater
from fjformer.jaxpruner.base_updater import NoPruning as NoPruning
from fjformer.jaxpruner.base_updater import SparseState as SparseState
from fjformer.jaxpruner.base_updater import apply_mask as apply_mask
from fjformer.jaxpruner.sparsity_schedules import NoUpdateSchedule as NoUpdateSchedule
from fjformer.jaxpruner.sparsity_schedules import OneShotSchedule as OneShotSchedule
from fjformer.jaxpruner.sparsity_schedules import PeriodicSchedule as PeriodicSchedule
from fjformer.jaxpruner.sparsity_schedules import (
	PolynomialSchedule as PolynomialSchedule,
)
from fjformer.jaxpruner.sparsity_types import SparsityType as SparsityType
from fjformer.jaxpruner.utils import summarize_intersection as summarize_intersection
from fjformer.jaxpruner.utils import summarize_sparsity as summarize_sparsity



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utility functions in utils.py."""

import jax.numpy as jnp
from absl.testing import absltest, parameterized

from fjformer.jaxpruner import utils


class UtilsTest(parameterized.TestCase, absltest.TestCase):
	def testSparsitySummary(self):
		param_tree = {
			"a": jnp.ones((2, 2)),
			"b": {
				"c": None,
				"d": jnp.array([1, 0, 1, 0.001]),
				"e": jnp.zeros([2, 1]),
			},
		}
		expected_summary = {
			"_total_sparsity": 0.3,
			"_nparams": 10.0,
			"_nparams_nnz": 7.0,
		}
		self.assertEqual(
			utils.summarize_sparsity(param_tree, only_total_sparsity=True),
			expected_summary,
		)

		expected_summary = {
			"_total_sparsity": 0.3,
			"_nparams": 10.0,
			"_nparams_nnz": 7.0,
			"a": jnp.array([0]),
			"b/c": None,
			"b/d": jnp.array([0.25]),
			"b/e": jnp.array([1.0]),
		}
		self.assertEqual(utils.summarize_sparsity(param_tree), expected_summary)

	def testInstersectionSummary(self):
		mask_tree = {
			"a": jnp.array([0, 0, 1, 1]),
			"b": {
				"c": None,
				"d": jnp.array([1, 0]),
				"e": jnp.array([1, 0]),
			},
		}
		second_mask_tree = {
			"a": jnp.array([0, 1, 0, 1]),
			"b": {
				"c": None,
				"d": jnp.array([1, 0]),
				"e": jnp.array([0, 1]),
			},
		}
		expected_summary = {"_total_intersection": 0.5}
		self.assertEqual(
			utils.summarize_intersection(
				mask_tree, second_mask_tree, only_total_intersection=True
			),
			expected_summary,
		)

		expected_summary = {
			"_total_intersection": 0.5,
			"a": jnp.array([0.5]),
			"b/c": None,
			"b/d": jnp.array([1.0]),
			"b/e": jnp.array([0.0]),
		}
		self.assertEqual(
			utils.summarize_intersection(
				mask_tree,
				second_mask_tree,
			),
			expected_summary,
		)


if __name__ == "__main__":
	absltest.main()
